from rdkit.Chem import MolSurf
from rdkit.Chem.EState import EState_VSA as RDKit_EState_VSA

from ._base import Descriptor

__all__ = (
    'LabuteASA',
    'PEOE_VSA', 'SMR_VSA', 'SlogP_VSA',
    'EState_VSA', 'VSA_EState',
)


class LabuteASA(Descriptor):
    r"""Labute's Approximate Surface Area descriptor(rdkit wrapper)."""
    __slots__ = ()
    explicit_hydrogens = False

    @classmethod
    def preset(cls):
        yield cls()

    def __str__(self):
        return self.__class__.__name__

    def parameters(self):
        return ()

    def calculate(self):
        return MolSurf.LabuteASA(self.mol)

    rtype = float


class MoeTypeBase(Descriptor):
    __slots__ = ('_k',)
    explicit_hydrogens = False
    _module = MolSurf

    @classmethod
    def preset(cls):
        return map(cls, range(1, cls.k_max))

    def __str__(self):
        return self.__class__.__name__ + str(self._k)

    def parameters(self):
        return self._k,

    def __init__(self, k=1):
        assert 1 <= k <= self.k_max
        self._k = k

    def calculate(self):
        f = getattr(self._module, str(self))
        return f(self.mol)

    rtype = float


class PEOE_VSA(MoeTypeBase):
    r"""MOE type descriptors using gasteiger charge and surface area contribution(rdkit wrapper).

    :type k: int
    :param k: (:math:`1 <= k <= k_{\rm max}`)
    """
    __slots__ = ()
    k_max = 14


class SMR_VSA(MoeTypeBase):
    r"""MOE type descriptors using Wildman-Crippen MR and surface area contribution(rdkit wrapper).

    :type k: int
    :param k: (:math:`1 <= k <= k_{\rm max}`)
    """
    __slots__ = ()
    k_max = 10


class SlogP_VSA(MoeTypeBase):
    r"""MOE type descriptors using Wildman-Crippen LogP and surface area contribution(rdkit wrapper).

    :type k: int
    :param k: (:math:`1 <= k <= k_{\rm max}`)
    """
    __slots__ = ()
    k_max = 12


class EState_VSA(MoeTypeBase):
    r"""MOE type descriptors using EState indices and surface area contribution(rdkit wrapper).

    :type k: int
    :param k: (:math:`1 <= k <= k_{\rm max}`)
    """
    __slots__ = ()
    _module = RDKit_EState_VSA
    k_max = 11


class VSA_EState(MoeTypeBase):
    r"""MOE type descriptors using EState indices and surface area contribution(rdkit wrapper).

    :type k: int
    :param k: (:math:`1 <= k <= k_{\rm max}`)
    """
    __slots__ = ()
    _module = RDKit_EState_VSA
    k_max = 10



from __future__ import division

from itertools import chain

import numpy as np

from ._graph_matrix import DistanceMatrix3D
from ._atomic_property import AtomicProperty
from ._base.descriptor import Descriptor

__all__ = 'MoRSE',


class MoRSE(Descriptor):
    __slots__ = ('_prop', '_distance')

    require_3D = True

    @classmethod
    def preset(cls):
        return chain(
            (cls(None, i) for i in range(1, 33)),
            (cls('m', i) for i in range(1, 33)),
            (cls('v', i) for i in range(1, 33)),
            (cls('se', i) for i in range(1, 33)),
            (cls('p', i) for i in range(1, 33)),
        )

    def __str__(self):
        p = '' if self._prop is None else self._prop.as_argument
        return 'Mor{:02d}{}'.format(self._distance, p)

    def parameters(self):
        p = None if self._prop is None else self._prop.as_argument
        return p, self._distance

    def __init__(self, prop=None, distance=2):
        if prop is None:
            self._prop = None
        else:
            self._prop = AtomicProperty(self.explicit_hydrogens, prop)

        self._distance = distance

    def dependencies(self):
        d = {'D': DistanceMatrix3D(self.explicit_hydrogens)}

        if self._prop is not None:
            d['A'] = self._prop

        return d

    def calculate(self, D, A=None):
        if D.shape[0] <= 1:
            self.fail(ValueError('require 2 or more atoms'))

        N = D.shape[0]

        if A is None:
            A = np.ones(N)
        else:
            A = A / self._prop.carbon

        A = A.reshape(1, -1)

        if self._distance == 1:
            n = np.ones((N, N), dtype='float')

        else:
            with self.rethrow_zerodiv():
                sr = (self._distance - 1) * D
                np.fill_diagonal(sr, 1)
                n = np.sin(sr) / sr

        np.fill_diagonal(n, 0)

        return np.float(0.5 * A.dot(n).dot(A.T))

    rtype = float



from rdkit.Chem import rdMolDescriptors

from ._base import Descriptor
from ._atomic_property import halogen

__all__ = (
    'AtomCount',
)


class AtomCount(Descriptor):
    r"""atom count descriptor.

    :type type: str
    :param type: type to count.

        * 'Atom'
        * 'HeavyAtom'
        * 'Spiro'
        * 'Bridgehead'
        * 'X' - all halogen
        * element symbol
    """
    __slots__ = ('_type',)

    @classmethod
    def preset(cls):
        return map(cls, [
            'Atom', 'HeavyAtom', 'Spiro', 'Bridgehead',
            'H', 'B', 'C', 'N', 'O', 'S', 'P', 'F', 'Cl', 'Br', 'I', 'X',
        ])

    @property
    def explicit_hydrogens(self):
        u"""require explicit_hydrogens when type is 'H' or 'Atom'."""
        return self._type in set(['H', 'Atom'])

    def __str__(self):
        return 'n' + self._type

    def parameters(self):
        return self._type,

    def __init__(self, type='Atom'):
        self._type = type

    def _calc_X(self):
        X = halogen
        return sum(a.GetAtomicNum() in X for a in self.mol.GetAtoms())

    def _calc(self):
        return sum(a.GetSymbol() == self._type for a in self.mol.GetAtoms())

    def _calc_all(self):
        return self.mol.GetNumAtoms()

    def calculate(self):
        if self._type == 'X':
            return self._calc_X()
        elif self._type in ['Atom', 'HeavyAtom']:
            return self._calc_all()
        elif self._type == 'Spiro':
            return rdMolDescriptors.CalcNumSpiroAtoms(self.mol)
        elif self._type == 'Bridgehead':
            return rdMolDescriptors.CalcNumBridgeheadAtoms(self.mol)
        else:
            return self._calc()

    rtype = int



import os
from glob import glob

import yaml
import numpy as np
from rdkit import Chem
from nose.tools import eq_
from numpy.testing import assert_almost_equal

from mordred import Calculator, Polarizability, descriptors
from mordred.error import MissingValueBase

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


data_dir = os.path.join(
    os.path.dirname(__file__),
    'references'
)


def test_by_references():
    calc = Calculator(
        d
        for d in descriptors.all
        if d.__class__ not in [Polarizability.APol, Polarizability.BPol]
    )

    calc.register([
        Polarizability.APol(True),
        Polarizability.BPol(True),
    ])

    actuals = dict()
    for mol in Chem.SDMolSupplier(os.path.join(data_dir, 'structures.sdf'), removeHs=False):
        actuals[mol.GetProp('_Name')] = {str(d): v for d, v in zip(calc.descriptors, calc(mol))}

    for path in glob(os.path.join(data_dir, '*.yaml')) + glob(os.path.join(data_dir, '**/*.yaml')):
        for test in yaml.load(open(path), Loader=Loader):
            dnames = test['names']
            if not isinstance(dnames, list):
                dnames = [dnames]

            desireds = (
                (mname, zip(dnames, values if isinstance(values, list) else [values]))
                for mname, values in test['results'].items()
            )

            digit = test.get('digit')
            if digit is None:
                assert_f = eq_
            else:
                def assert_f(a, d, m):
                    if np.isnan(d):
                        assert isinstance(a, MissingValueBase)
                        return

                    assert_almost_equal(a, d, digit, m)

            for mname, descs in desireds:
                for dname, desired in descs:
                    if not desired == 'skip':
                        yield (
                            assert_f,
                            actuals[mname][dname],
                            desired,
                            '{} of {}'.format(dname, mname)
                        )



from six import string_types, integer_types
from numpy import product

from ._base import Descriptor
from ._graph_matrix import Valence, DistanceMatrix
from ._atomic_property import table

__all__ = ('MolecularDistanceEdge',)


class MolecularDistanceEdge(Descriptor):
    r"""molecular distance edge descriptor.

    :type valence1: int
    :param valence1: valence of first atom

    :type valence2: int
    :param valence2: valence of second atom

    :type element: str or int
    :param element: atomic symbol or atomic number

    :returns: NaN when :math:`N_{\rm MDE} = 0`
    """
    __slots__ = ('_valence1', '_valence2', '_atomic_num',)
    explicit_hydrogens = False

    @classmethod
    def preset(cls):
        return (
            cls(a, b, e)
            for e in [6, 8, 7]
            for a in range(1, 11 - e)
            for b in range(a, 11 - e)
        )

    def __str__(self):
        return 'MDE{}-{}{}'.format(
            table.GetElementSymbol(self._atomic_num),
            self._valence1,
            self._valence2,
        )

    def parameters(self):
        return self._valence1, self._valence2, self._atomic_num

    def __init__(self, valence1=1, valence2=1, element='C'):
        self._valence1 = min(valence1, valence2)
        self._valence2 = max(valence1, valence2)
        if isinstance(element, integer_types):
            self._atomic_num = element
        elif isinstance(element, string_types):
            self._atomic_num = table.GetAtomicNumber(element)
        else:
            raise ValueError('element must be atomic number or atomic symbol')

    def dependencies(self):
        return {
            'D': DistanceMatrix(self.explicit_hydrogens),
            'V': Valence(self.explicit_hydrogens),
        }

    def calculate(self, D, V):
        N = self.mol.GetNumAtoms()
        Dv = [
            D[i, j]
            for i in range(N)
            for j in range(i + 1, N)
            if (V[i] == self._valence1 and V[j] == self._valence2) or
            (V[j] == self._valence1 and V[i] == self._valence2)
            if self.mol.GetAtomWithIdx(i).GetAtomicNum() ==
            self.mol.GetAtomWithIdx(j).GetAtomicNum() ==
            self._atomic_num
        ]
        n = len(Dv)

        with self.rethrow_zerodiv():
            dx = product(Dv) ** (1. / (2. * n))

        return n / (dx ** 2)

    rtype = float



from __future__ import print_function

import os
import sys

import numpy as np
from tqdm import tqdm_notebook


def parse_enum(enum, v):
    if isinstance(v, enum):
        return v
    else:
        return enum[v]


def atoms_to_numpy(f, mol, dtype='float'):
    return np.fromiter(
        (f(a) for a in mol.GetAtoms()),
        dtype, mol.GetNumAtoms()
    )


def conformer_to_numpy(conf):
    return np.array(
        [list(conf.GetAtomPosition(i)) for i in range(conf.GetNumAtoms())]
    )


class Capture(object):
    def __init__(self, target='stderr'):
        self.target = target
        self.orig = getattr(sys, target)
        self.result = []

    def write(self, text):
        self.result.append(text)

    def flush(self):
        pass

    def __enter__(self):
        setattr(sys, self.target, self)
        return self

    def __exit__(self, *args):
        setattr(sys, self.target, self.orig)


class DummyBar(object):
    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args, **kws):
        pass

    def update(self, *args, **kws):
        pass

    @classmethod
    def write(cls, s, file=sys.stdout, end='\n'):
        print(s, file=file, end=end)


class NotebookWrapper(object):
    def __init__(self, **kwargs):
        self.bar = tqdm_notebook(**kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def update(self, *args, **kwargs):
        self.bar.update(*args, **kwargs)

    def write(self, *args, **kwargs):
        self.bar.update(*args, **kwargs)


def PathType(string):
    if not os.path.isfile(string):
        raise ValueError('file not exists: {}'.format(string))

    return string


def module_prog(pkg):
    return '{} -m {}'.format(os.path.basename(sys.executable), pkg)



import numpy as np

from ._base import Descriptor
from ._graph_matrix import DistanceMatrix
from ._atomic_property import AtomicProperty, get_properties

__all__ = ('ATS', 'AATS', 'ATSC', 'AATSC', 'MATS', 'GATS',)


class AutocorrelationBase(Descriptor):
    __slots__ = '_prop', '_order'
    explicit_hydrogens = True

    def __str__(self):
        return '{}{}{}'.format(
            self.__class__.__name__,
            self._order,
            self._avec.as_argument
        )

    def parameters(self):
        return self._order, self._prop

    def __init__(self, order=0, prop='m'):
        self._prop = prop
        self._order = order

    @property
    def _avec(self):
        return AtomicProperty(self.explicit_hydrogens, self._prop)

    @property
    def _cavec(self):
        return CAVec(self._prop)

    @property
    def _gmat(self):
        return GMat(self._order)

    @property
    def _gsum(self):
        return GSum(self._order)

    @property
    def _ATS(self):
        return ATS(self._order, self._prop)

    @property
    def _ATSC(self):
        return ATSC(self._order, self._prop)

    @property
    def _AATSC(self):
        return AATSC(self._order, self._prop)

    rtype = float


class AutocorrelationProp(AutocorrelationBase):
    __slots__ = ()

    def parameters(self):
        return self._prop,

    def __init__(self, prop):
        self._prop = prop

    rtype = None


class AutocorrelationOrder(AutocorrelationBase):
    __slots__ = ()

    def _prop(self):
        pass

    def parameters(self):
        return self._order,

    def __init__(self, order):
        self._order = order

    rtype = None


class CAVec(AutocorrelationProp):
    __slots__ = ()
    _order = 0

    def dependencies(self):
        return {'avec': self._avec}

    def calculate(self, avec):
        return avec - avec.mean()


class GMat(AutocorrelationOrder):
    __slots__ = ()

    def dependencies(self):
        return {'dmat': DistanceMatrix(self.explicit_hydrogens)}

    def calculate(self, dmat):
        return dmat == self._order


class GSum(AutocorrelationOrder):
    __slots__ = ()

    def dependencies(self):
        return {'gmat': GMat(self._order)}

    def calculate(self, gmat):
        s = gmat.sum()

        return s if self._order == 0 else 0.5 * s


MAX_DISTANCE = 8


class ATS(AutocorrelationBase):
    r"""Autocorrelation of Topological Structure descriptor.

    a.k.a. Moreau-Broto autocorrelation descriptor

    .. math::
        {\rm ATS}_0 = \sum^{A}_{i=1} {\boldsymbol w}_i^2

        {\rm ATS}_k = \frac{1}{2}
            \left(
                {\boldsymbol w}^{\rm T} \cdot
                {}^k{\boldsymbol B} \cdot
                {\boldsymbol w}
            \right)

        {}^k{\boldsymbol B} =
            \begin{cases}
                1 & (d_{ij} =    k) \\
                0 & (d_{ij} \neq k)
            \end{cases}

    where
    :math:`{\boldsymbol w}` is atomic property vector,
    :math:`d_{ij}` is graph distance(smallest number of bonds between atom i and j).

    :type order: int
    :param order: order(:math:`k`)

    :type property: str, function
    :param property: :ref:`atomic_properties`

    :returns: NaN when any properties are NaN
    """

    __slots__ = ()

    @classmethod
    def preset(cls):
        return (
            cls(d, a)
            for a in get_properties(valence=True)
            for d in range(MAX_DISTANCE + 1)
        )

    def dependencies(self):
        return {'avec': self._avec, 'gmat': self._gmat}

    def calculate(self, avec, gmat):
        if self._order == 0:
            return (avec ** 2).sum().astype('float')

        return 0.5 * avec.dot(gmat).dot(avec)


class AATS(ATS):
    r"""averaged ATS descriptor.

    .. math::

        {\rm AATS}_k = \frac{{\rm ATS}_k}{\Delta_k}

    where
    :math:`\Delta_k` is number of vertex pairs at order equal to :math:`k`.

    :Parameters: see :py:class:`ATS`

    :returns: NaN when

        * :math:`\Delta_k = 0`
        * some properties are NaN
    """

    __slots__ = ()

    def dependencies(self):
        return {'ATS': self._ATS, 'gsum': self._gsum}

    def calculate(self, ATS, gsum):
        with self.rethrow_zerodiv():
            return ATS / gsum


class ATSC(AutocorrelationBase):
    r"""centered ATS descriptor.

    ATS with :math:`{\boldsymbol w}_{\rm c}` property

    .. math::
        {\boldsymbol w}_{\rm c} = {\boldsymbol w} - \bar{\boldsymbol w}

    :Parameters: see :py:class:`ATS`

    :returns: NaN when any properties are NaN
    """

    __slots__ = ()

    @classmethod
    def preset(cls):
        return (
            cls(d, a)
            for a in get_properties(charge=True, valence=True)
            for d in range(MAX_DISTANCE + 1)
        )

    def dependencies(self):
        return {'cavec': self._cavec, 'gmat': self._gmat}

    def calculate(self, cavec, gmat):
        if self._order == 0:
            return (cavec ** 2).sum().astype('float')

        return 0.5 * cavec.dot(gmat).dot(cavec)


class AATSC(ATSC):
    r"""averaged ATSC descriptor.

    .. math::

        {\rm AATSC}_k = \frac{{\rm ATSC}_k}{\Delta_k}

    where
    :math:`\Delta_k` is number of vertex pairs at order equal to :math:`k`.

    :Parameters: see :py:class:`ATS`

    :returns: NaN when

        * :math:`\Delta_k = 0`
        * any properties are NaN
    """

    __slots__ = ()

    def dependencies(self):
        return {'ATSC': self._ATSC, 'gsum': self._gsum}

    def calculate(self, ATSC, gsum):
        with self.rethrow_zerodiv():
            return ATSC / gsum


class MATS(AutocorrelationBase):
    r"""Moran coefficient descriptor.

    .. math::

        {\rm MATS}_k = \frac{
            {\rm AATSC}_k
            }{
            \frac{1}{A} \cdot \sum {\boldsymbol w}_{\rm c}^2
            }

    :Parameters: see :py:class:`ATS`

    :returns: NaN when

        * some properties are NaN
        * denominator = 0
    """

    __slots__ = ()

    @classmethod
    def preset(cls):
        return (
            cls(d, a)
            for a in get_properties(charge=True, valence=True)
            for d in range(1, MAX_DISTANCE + 1)
        )

    def dependencies(self):
        return {
            'avec': self._avec,
            'AATSC': self._AATSC,
            'cavec': self._cavec,
        }

    def calculate(self, avec, AATSC, cavec):
        with self.rethrow_zerodiv():
            return len(avec) * AATSC / (cavec ** 2).sum()


class GATS(MATS):
    r"""Geary coefficient descriptor.

    :Parameters: see :py:class:`ATS`

    :returns: NaN when

        * :math:`\Delta_k = 0`
        * any properties are NaN
        * denominator = 0
    """

    __slots__ = ()

    def dependencies(self):
        return {
            'avec': self._avec,
            'gmat': self._gmat,
            'gsum': self._gsum,
            'cavec': self._cavec,
        }

    def calculate(self, avec, gmat, gsum, cavec):
        if len(avec) <= 1:
            self.fail(ValueError('no bond'))

        with self.rethrow_zerodiv():
            n = (gmat * (avec[:, np.newaxis] - avec) ** 2).sum() / (4 * gsum)
            d = (cavec ** 2).sum() / (len(avec) - 1)
            return n / d



import inspect
import operator
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager

import six
import numpy as np


class MissingValueException(Exception):
    "internally used exception"

    __slots__ = ('error',)

    def __init__(self, error):
        self.error = error


class Descriptor(six.with_metaclass(ABCMeta, object)):
    r"""abstract base class of descriptors.

    Attributes:
        mol(rdkit.Mol): target molecule
    """

    __slots__ = '_context',

    explicit_hydrogens = True
    kekulize = False
    require_connected = False
    require_3D = False

    def __reduce_ex__(self, version):
        return self.__class__, self.parameters()

    @classmethod
    def preset(cls):
        r"""generate preset descriptor instances.

        Returns:
            Iterable[Descriptor]: preset descriptors
        """
        return ()

    @abstractmethod
    def parameters(self):
        '''[abstractmethod] get __init__ arguments of this descriptor instance.

        this method used in pickling and identifying descriptor instance.

        Returns:
            tuple: tuple of __init__ arguments
        '''
        raise NotImplementedError('not implemented Descriptor.parameters method')

    @abstractmethod
    def calculate(self):
        r"""[abstractmethod] calculate descriptor value.

        Returns:
            rtype
        """
        raise NotImplementedError('not implemented Descriptor.calculate method')

    def dependencies(self):
        r"""descriptor dependencies.

        Returns:
            dict[str, Descriptor or None] or None
        """
        pass

    @property
    def as_argument(self):
        '''argument representation of descriptor

        Returns:
            any
        '''
        return self

    @staticmethod
    def _pretty(v):
        v = getattr(v, 'as_argument', v)
        return repr(v)

    def __repr__(self):
        return '{}.{}({})'.format(
            self.__class__.__module__,
            self.__class__.__name__,
            ', '.join(self._pretty(a) for a in self.parameters())
        )

    def __hash__(self):
        return hash((self.__class__, self.parameters()))

    def __compare_by_reduce(meth):
        def compare(self, other):
            l = self.__class__, self.parameters()
            r = other.__class__, other.parameters()
            return getattr(l, meth)(r)

        return compare

    __eq__ = __compare_by_reduce('__eq__')
    __ne__ = __compare_by_reduce('__ne__')

    __lt__ = __compare_by_reduce('__lt__')
    __gt__ = __compare_by_reduce('__gt__')
    __le__ = __compare_by_reduce('__le__')
    __ge__ = __compare_by_reduce('__ge__')

    rtype = None

    @property
    def mol(self):
        '''get molecule

        Returns:
            rdkit.Mol
        '''
        return self._context.get_mol(self)

    @property
    def coord(self):
        '''get 3D coordinate

        Returns:
            numpy.array[3, N]: coordinate matrix
        '''
        if not self.require_3D:
            self.fail(AttributeError('use 3D coordinate in 2D descriptor'))

        return self._context.get_coord(self)

    def fail(self, exception):
        '''raise known exception and return missing value

        Raises:
            MissingValueException
        '''
        raise MissingValueException(exception)

    @contextmanager
    def rethrow_zerodiv(self):
        '''[contextmanager] treat zero div as known exception
        '''

        with np.errstate(divide='raise', invalid='raise'):
            try:
                yield
            except (FloatingPointError, ZeroDivisionError) as e:
                self.fail(ZeroDivisionError(*e.args))

    @contextmanager
    def rethrow_na(self, exception):
        '''[contextmanager] treat any exceptions as known exception
        '''

        try:
            yield
        except exception as e:
            self.fail(e)

    def _unary_common(name, operator):
        def unary(self):
            return UnaryOperatingDescriptor(name.format(self), operator, self)

        return unary

    def _binary_common(name, operator):
        def binary(self, other):
            if not isinstance(other, Descriptor):
                other = ConstDescriptor(str(other), other)

            return BinaryOperatingDescriptor(name.format(self, other), operator, self, other)

        return binary

    __add__ = _binary_common('({}+{})', operator.add)
    __sub__ = _binary_common('({}-{})', operator.sub)
    __mul__ = _binary_common('({}*{})', operator.mul)
    __truediv__ = _binary_common('({}/{})', operator.truediv)
    __floordiv__ = _binary_common('({}//{})', operator.floordiv)
    __mod__ = _binary_common('({}%{})', operator.mod)
    __pow__ = _binary_common('({}**{})', operator.pow)

    __neg__ = _unary_common('-{}', operator.neg)
    __pos__ = _unary_common('+{}', operator.pos)
    __abs__ = _unary_common('|{}|', operator.abs)

    if six.PY3:
        __ceil__ = _unary_common('ceil({})', np.ceil)
        __floor__ = _unary_common('floor({})', np.floor)

    __trunc__ = _unary_common('trunc({})', np.trunc)


def is_descriptor_class(desc):
    r"""check calculatable descriptor class or not.

    Returns:
        bool
    """
    return (
        isinstance(desc, type) and
        issubclass(desc, Descriptor) and
        not inspect.isabstract(desc)
    )


class UnaryOperatingDescriptor(Descriptor):
    @classmethod
    def preset(cls):
        return cls()

    def parameters(self):
        return self.name, self.operator, self.value

    def __init__(self, name, operator, value):
        self.name = name
        self.operator = operator
        self.value = value

    def __str__(self):
        return self.name

    def dependencies(self):
        return {
            'value': self.value,
        }

    def calculate(self, value):
        return self.operator(value)


class ConstDescriptor(Descriptor):
    @classmethod
    def preset(cls):
        return cls()

    def parameters(self):
        return self.name, self.value

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __str__(self):
        return self.name

    def calculate(self):
        return self.value


class BinaryOperatingDescriptor(Descriptor):
    @classmethod
    def preset(cls):
        return cls()

    def parameters(self):
        return self.name, self.operator, self.left, self.right

    def __init__(self, name, operator, left, right):
        self.name = name
        self.operator = operator
        self.left = left
        self.right = right

    def __str__(self):
        return self.name

    def dependencies(self):
        return {
            'left': self.left,
            'right': self.right,
        }

    def calculate(self, left, right):
        return self.operator(left, right)



import numpy as np

from ._base import Descriptor
from ._util import atoms_to_numpy
from ._graph_matrix import AdjacencyMatrix, DistanceMatrix3D

__all__ = ('GravitationalIndex',)


class GravitationalIndex(Descriptor):
    __slots__ = '_heavy', '_pair'

    @classmethod
    def preset(cls):
        return (
            cls(h, p)
            for p in [False, True]
            for h in [True, False]
        )

    require_3D = True

    @property
    def explicit_hydrogens(self):
        return not self._heavy

    def __str__(self):
        return 'GRAV{}{}'.format(
            '' if self._heavy else 'H',
            'p' if self._pair else '',
        )

    def parameters(self):
        return self._heavy, self._pair

    def __init__(self, heavy=True, pair=False):
        self._heavy = heavy
        self._pair = pair

    def dependencies(self):
        d = {'D': DistanceMatrix3D(self.explicit_hydrogens)}

        if self._pair:
            d['A'] = AdjacencyMatrix(self.explicit_hydrogens)

        return d

    def calculate(self, D, A=1.0):
        w = atoms_to_numpy(lambda a: a.GetMass(), self.mol)

        w = w[:, np.newaxis] * w
        np.fill_diagonal(w, 0)

        D = D.copy()
        np.fill_diagonal(D, 1)

        with self.rethrow_zerodiv():
            return 0.5 * np.sum(w * A / D ** 2)

    rtype = float



import numpy as np

from ._base import Descriptor
from ._atomic_property import AtomicProperty, get_properties

__all__ = ('BCUT',)


class BCUTBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False
    require_connected = True


class Burden(BCUTBase):
    __slots__ = ()

    def parameters(self):
        return ()

    def calculate(self):
        N = self.mol.GetNumAtoms()

        mat = 0.001 * np.ones((N, N))

        for bond in self.mol.GetBonds():
            a = bond.GetBeginAtom()
            b = bond.GetEndAtom()
            i = a.GetIdx()
            j = b.GetIdx()

            try:
                w = bond.GetBondTypeAsDouble() / 10.0
            except RuntimeError:
                self.fail(ValueError('unknown bond type'))

            if a.GetDegree() == 1 or b.GetDegree() == 1:
                w += 0.01

            mat[i, j] = w
            mat[j, i] = w

        return mat


class BurdenEigenValues(BCUTBase):
    __slots__ = ('_prop',)

    def parameters(self):
        return self._prop,

    def __init__(self, prop):
        self._prop = prop

    def dependencies(self):
        return {
            'ps': self._prop,
            'burden': Burden(),
        }

    def calculate(self, burden, ps):
        bmat = burden.copy()

        np.fill_diagonal(bmat, ps)
        ev = np.linalg.eig(bmat)[0]

        if np.iscomplexobj(ev):
            ev = ev.real

        return np.sort(ev)[-1::-1]


class BCUT(BCUTBase):
    r"""BCUT descriptor.

    :type prop: :py:class:`str` or :py:class:`function`
    :param prop: :ref:`atomic_properties`

    :type nth: int
    :param nth: n-th eigen value. 0 is highest, -1 is lowest.

    :returns: NaN when

        * any atomic properties are NaN
        * :math:`\left| nth \right| > A`
    """
    __slots__ = ('_prop', '_nth',)

    @classmethod
    def preset(cls):
        return (
            cls(a, n)
            for a in get_properties(valence=True, charge=True)
            for n in [0, -1]
        )

    def __str__(self):
        if self._nth < 0:
            return 'BCUT{}-{}l'.format(self._prop.as_argument, np.abs(self._nth))
        else:
            return 'BCUT{}-{}h'.format(self._prop.as_argument, self._nth + 1)

    def parameters(self):
        return self._prop, self._nth

    def __init__(self, prop='m', nth=0):
        self._prop = AtomicProperty(self.explicit_hydrogens, prop)
        self._nth = nth

    def dependencies(self):
        return {'bev': BurdenEigenValues(self._prop)}

    def calculate(self, bev):
        try:
            return bev[self._nth]
        except IndexError:
            self.fail(ValueError('nth greater then atom count'))

    rtype = float



from __future__ import division

from collections import defaultdict

from rdkit.Chem import HybridizationType

from ._base import Descriptor

__all__ = (
    'CarbonTypes', 'HybridizationRatio',
)


class CarbonTypesBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False
    kekulize = True


class CarbonTypesCache(CarbonTypesBase):
    __slots__ = ()

    def parameters(self):
        return ()

    _hybridization = {
        HybridizationType.SP: 1,
        HybridizationType.SP2: 2,
        HybridizationType.SP3: 3,
        HybridizationType.SP3D: 3,
        HybridizationType.SP3D2: 3,
    }

    def calculate(self):
        r = defaultdict(lambda: defaultdict(int))
        for a in self.mol.GetAtoms():
            if a.GetAtomicNum() != 6:
                continue

            carbon = sum(
                other.GetAtomicNum() == 6
                for other in a.GetNeighbors()
            )

            SP = self._hybridization.get(a.GetHybridization())

            r[SP][carbon] += 1

        return r


class CarbonTypes(CarbonTypesBase):
    r"""carbon types descriptor.

    :type nCarbon: int
    :param nCarbon: count `n`-carbon bonded carbon

    :type SP: int
    :param SP: count :math:`{\rm SP}n` carbon
    """
    __slots__ = ('_nCarbon', '_SP',)

    @classmethod
    def preset(cls):
        return map(lambda args: cls(*args), [
            (1, 1), (2, 1),
            (1, 2), (2, 2), (3, 2),
            (1, 3), (2, 3), (3, 3), (4, 3),
        ])

    def __str__(self):
        return 'C{}SP{}'.format(self._nCarbon, self._SP)

    def parameters(self):
        return self._nCarbon, self._SP

    def __init__(self, nCarbon=1, SP=3):
        assert SP in [1, 2, 3]

        self._nCarbon = nCarbon
        self._SP = SP

    def dependencies(self):
        return {'CT': CarbonTypesCache()}

    def calculate(self, CT):
        return CT[self._SP][self._nCarbon]

    rtype = int


class HybridizationRatio(CarbonTypesBase):
    r"""hybridization ratio descriptor.

    .. math::

        {\rm HybRatio} = \frac{N_{\rm SP3}}{N_{\rm SP2} + N_{\rm SP3}}

    :returns: NaN when :math:`N_{\rm SP2} + N_{\rm SP3} = 0`.
    """
    __slots__ = ()

    @classmethod
    def preset(cls):
        yield cls()

    def __str__(self):
        return 'HybRatio'

    def parameters(self):
        return ()

    def dependencies(self):
        return {'CT': CarbonTypesCache()}

    def calculate(self, CT):
        Nsp3 = sum(CT[3].values())
        Nsp2 = sum(CT[2].values())

        if Nsp3 == Nsp2 == 0:
            self.fail(ValueError('there are no sp3 and sp2 carbons'))

        return Nsp3 / (Nsp2 + Nsp3)

    rtype = float



import numpy as np


# http://prideout.net/blog/?p=44
# http://blog.andreaskahler.com/2009/06/creating-icosphere-mesh-in-code.html


class SphereMesh(object):
    def __init__(self, level=4):
        t = (1.0 + np.sqrt(5.0)) / 2.0
        self.vertices = np.array([
            (-1, t, 0),
            (1, t, 0),
            (-1, -t, 0),
            (1, -t, 0),

            (0, -1, t),
            (0, 1, t),
            (0, -1, -t),
            (0, 1, -t),

            (t, 0, -1),
            (t, 0, 1),
            (-t, 0, -1),
            (-t, 0, 1),
        ], dtype='float')

        self.faces = np.array([
            (0, 11, 5),
            (0, 5, 1),
            (0, 1, 7),
            (0, 7, 10),
            (0, 10, 11),

            (1, 5, 9),
            (5, 11, 4),
            (11, 10, 2),
            (10, 7, 6),
            (7, 1, 8),

            (3, 9, 4),
            (3, 4, 2),
            (3, 2, 6),
            (3, 6, 8),
            (3, 8, 9),

            (4, 9, 5),
            (2, 4, 11),
            (6, 2, 10),
            (8, 6, 7),
            (9, 8, 1),
        ], dtype='int')

        self.normalize(0)

        self.level = 1
        self.subdivide(level)

    def normalize(self, begin):
        self.vertices[begin:] /= np.sqrt((self.vertices[begin:] ** 2).sum(axis=1))[:, np.newaxis]

    def _subdivide(self):
        self.level += 1

        Nv = len(self.vertices)
        Nf = len(self.faces)

        A = self.faces[:, 0]
        B = self.faces[:, 1]
        C = self.faces[:, 2]

        Av = self.vertices[A]
        Bv = self.vertices[B]
        Cv = self.vertices[C]

        self.vertices = np.r_[
            self.vertices,
            Av + Bv,
            Bv + Cv,
            Av + Cv
        ]
        self.normalize(Nv)

        AB = np.arange(len(self.faces)) + Nv
        BC = AB + Nf
        AC = AB + 2 * Nf

        self.faces = np.concatenate([
            A, B, C, AB,
            AB, AB, AC, AC,
            AC, BC, BC, BC
        ]).reshape(3, -1).T

    def subdivide(self, level):
        for _ in range(level - self.level):
            self._subdivide()

    def __len__(self):
        return len(self.vertices)



import math
import operator

import six
from rdkit import Chem
from nose.tools import eq_

from .. import ABCIndex

binary = [
    operator.add,
    operator.sub,
    operator.mul,
    operator.truediv,
    operator.floordiv,
    operator.mod,
    operator.pow,
]

unary = [
    operator.neg,
    operator.pos,
    operator.abs,

    math.trunc,
]


if six.PY3:
    unary.extend([
        math.ceil,
        math.floor,
    ])


def test_compose_descriptor():
    l = ABCIndex.ABCIndex()
    r = ABCIndex.ABCGGIndex()
    m = Chem.MolFromSmiles('c1ccccc1C')

    for op in binary:
        yield eq_, op(l, r)(m), op(l(m), r(m))

    for op in unary:
        yield eq_, op(l)(m), op(l(m))



from __future__ import division

from ._base import Descriptor
from ._graph_matrix import Radius3D as CRadius3D
from ._graph_matrix import Diameter3D as CDiameter3D

__all__ = ('Diameter3D', 'Radius3D', 'GeometricalShapeIndex', 'PetitjeanIndex3D',)


class GeometricalIndexBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = True
    require_3D = True

    @classmethod
    def preset(cls):
        yield cls()

    def parameters(self):
        return ()

    rtype = float


class Radius3D(GeometricalIndexBase):
    r"""geometric radius descriptor."""
    __slots__ = ()

    def __str__(self):
        return 'GeomRadius'

    def dependencies(self):
        return {'R': CRadius3D(self.explicit_hydrogens)}

    def calculate(self, R):
        return R


class Diameter3D(GeometricalIndexBase):
    r"""geometric diameter descriptor."""
    __slots__ = ()

    def __str__(self):
        return 'GeomDiameter'

    def dependencies(self):
        return {'D': CDiameter3D(self.explicit_hydrogens)}

    def calculate(self, D):
        return D


class GeometricalShapeIndex(GeometricalIndexBase):
    r"""geometrical shape index descriptor.

    .. math::

        I_{\rm topo} = \frac{D - R}{R}

    where
    :math:`R` is geometric radius,
    :math:`D` is geometric diameter.

    :returns: NaN when :math:`R = 0`
    """
    __slots__ = ()

    def __str__(self):
        return 'GeomShapeIndex'

    def dependencies(self):
        return {
            'R': CRadius3D(self.explicit_hydrogens),
            'D': CDiameter3D(self.explicit_hydrogens),
        }

    def calculate(self, R, D):
        with self.rethrow_zerodiv():
            return (D - R) / (R)


class PetitjeanIndex3D(GeometricalShapeIndex):
    r"""geometric Petitjean index descriptor.

    .. math::

        I_{\rm Petitjean} = \frac{D - R}{D}

    where
    :math:`R` is geometric radius,
    :math:`D` is geometric diameter.

    :returns: NaN when :math:`D = 0`
    """
    __slots__ = ()

    def __str__(self):
        return 'GeomPetitjeanIndex'

    def calculate(self, R, D):
        with self.rethrow_zerodiv():
            return (D - R) / (D)



from ._base import Descriptor
from ._graph_matrix import DistanceMatrix

__all__ = ('WienerIndex',)


class WienerIndex(Descriptor):
    r"""Wiener index.

    :type polarity: bool
    :param polarity: use polarity Wiener index
    """
    __slots__ = ('_polarity',)
    explicit_hydrogens = False

    @classmethod
    def preset(cls):
        yield cls(False)
        yield cls(True)

    def __str__(self):
        return 'WPol' if self._polarity else 'WPath'

    def parameters(self):
        return self._polarity,

    def __init__(self, polarity=False):
        self._polarity = polarity

    def dependencies(self):
        return {'D': DistanceMatrix(self.explicit_hydrogens)}

    def calculate(self, D):
        if self._polarity:
            return int(0.5 * (D == 3).sum())
        else:
            return int(0.5 * D.sum())

    rtype = int



import os
import sys
import shutil
import tempfile
from contextlib import contextmanager

from nose.tools import eq_
from rdkit.Chem import AllChem as Chem

from .. import Calculator, __version__, descriptors
from ..__main__ import main as mordred

Nd2D = len(Calculator(descriptors, ignore_3D=True).descriptors)
Nd3D = len(Calculator(descriptors, ignore_3D=False).descriptors)


def in_(a, s):
    assert a in s, '{!r} not in {!r}'.format(a, s)


@contextmanager
def isolate():
    cwd = os.getcwd()
    try:
        work = tempfile.mkdtemp()
        os.chdir(work)
        yield
    finally:
        os.chdir(cwd)
        shutil.rmtree(work)


class Result(object):
    def __iter__(self):
        return iter([self.stdout, self.stderr, self.exitcode])

    def __repr__(self):
        return '{!r} {!r} {!r}'.format(self.stdout, self.stderr, self.exitcode)


def command(cmd, *args):
    r = Result()

    orig = sys.stdout, sys.stderr

    ofd, opath = tempfile.mkstemp()
    efd, epath = tempfile.mkstemp()

    try:
        with os.fdopen(ofd, 'w') as stdout, os.fdopen(efd, 'w') as stderr:
            sys.stdout = stdout
            sys.stderr = stderr

            try:
                cmd(args=args)
                r.exitcode = 0
            except SystemExit as e:
                r.exitcode = e.args[0]

    finally:
        sys.stdout, sys.stderr = orig

        with open(opath) as stdout, open(epath) as stderr:
            r.stdout = stdout.read()
            r.stderr = stderr.read()

        os.remove(opath)
        os.remove(epath)

    return r


def test_no_args():
    stdout, stderr, exitcode = command(mordred)
    eq_(exitcode, 2)
    eq_(stdout, '')
    in_('usage:', stderr)
    # python3 or python2
    assert 'the following arguments are required: INPUT' in stderr or 'too few arguments' in stderr


def test_help():
    stdout, stderr, exitcode = command(mordred, '-h')
    eq_(exitcode, 0)
    eq_(stderr, '')
    in_('usage:', stdout)
    in_('descriptors:', stdout)


def test_version():
    stdout, stderr, exitcode = command(mordred, '--version')
    eq_(exitcode, 0)

    vstr = 'mordred-{}\n'.format(__version__)

    if stderr == '':  # python 3
        eq_(stdout, vstr)
    else:  # python2
        eq_(stderr, vstr)


def test_missing_file():
    with isolate():
        stdout, stderr, exitcode = command(mordred, 'missing.smi')
        eq_(exitcode, 2)
        eq_(stdout, '')
        in_('usage:', stderr)
        in_('invalid PathType value', stderr)


def number_of_field(output):
    return len(list(filter(lambda c: c == ',', output.split('\n')[0])))


def test_smi():
    with isolate():
        with open('input.smi', 'w') as f:
            f.write('c1ccccc1 Benzene\n')

        stdout, stderr, exitcode = command(mordred, 'input.smi', '-q', '-o', '-')

        eq_(exitcode, 0)
        eq_(number_of_field(stdout), Nd2D)
        eq_(stdout.split('\n')[1].split(',')[0], 'Benzene')
        eq_(stderr, '')


def test_smi_without_name():
    with isolate():
        with open('input.smi', 'w') as f:
            f.write('c1ccccc1\n')

        stdout, stderr, exitcode = command(mordred, 'input.smi', '-q', '-o', '-')

        eq_(exitcode, 0)
        eq_(number_of_field(stdout), Nd2D)
        eq_(stdout.split('\n')[1].split(',')[0], 'c1ccccc1')
        eq_(stderr, '')


def test_sdf():
    with isolate():
        mol = Chem.MolFromSmiles('c1ccccc1')
        mol.SetProp('_Name', 'Benzene')
        Chem.MolToMolFile(mol, 'input.sdf')

        stdout, stderr, exitcode = command(mordred, 'input.sdf', '-q', '-o', 'output.csv')

        eq_(exitcode, 0)
        eq_(stdout, '')
        eq_(stderr, '')
        output = open('output.csv').read()
        eq_(number_of_field(output), Nd2D)
        eq_(output.split('\n')[1].split(',')[0], 'Benzene')


def test_sdf_3D():
    with isolate():
        mol = Chem.MolFromSmiles('c1ccccc1')
        Chem.EmbedMolecule(mol)
        mol.SetProp('_Name', 'Benzene')
        Chem.MolToMolFile(mol, 'input.sdf')

        stdout, stderr, exitcode = command(mordred, 'input.sdf', '-q', '-o', 'output.csv', '-3')

        eq_(exitcode, 0)
        eq_(stdout, '')
        eq_(stderr, '')
        output = open('output.csv').read()
        eq_(number_of_field(output), Nd3D)
        eq_(output.split('\n')[1].split(',')[0], 'Benzene')


def test_verbose():
    with isolate():
        with open('input.smi', 'w') as f:
            f.write('c1ccccc1 Benzene\n')

        r = command(mordred, 'input.smi', '-o', '-', '-q', '-vv')

        eq_(r.exitcode, 0)
        eq_(number_of_field(r.stdout), Nd2D)
        eq_(r.stdout.split('\n')[1].split(',')[0], 'Benzene')
        in_('[Missing]', r.stderr)



from ._base import Descriptor
from ._atomic_property import polarizability78, polarizability94

__all__ = ('APol', 'BPol',)


class PolarizabilityBase(Descriptor):
    __slots__ = '_use78',

    @classmethod
    def preset(cls):
        yield cls()

    def __str__(self):
        return self.__class__.__name__.lower() + ('78' if self._use78 else '')

    def parameters(self):
        return self._use78,

    def __init__(self, use78=False):
        self._use78 = use78

    def _get_table(self):
        return polarizability78 if self._use78 else polarizability94

    rtype = float


class APol(PolarizabilityBase):
    r"""atomic polarizability descriptor.

    :type use78: bool
    :param use78: use old atomic polarizability data
    """
    __slots__ = ()

    def calculate(self):
        table = self._get_table()
        return sum(table[a.GetAtomicNum()] for a in self.mol.GetAtoms())


class BPol(PolarizabilityBase):
    r"""bond polarizability descriptor.

    :type use78: bool
    :param use78: use old atomic polarizability data
    """
    __slots__ = ()

    def calculate(self):
        table = self._get_table()

        def bond_pol(bond):
            a = bond.GetBeginAtom().GetAtomicNum()
            b = bond.GetEndAtom().GetAtomicNum()
            return abs(table[a] - table[b])

        return float(sum(bond_pol(b) for b in self.mol.GetBonds()))



from enum import IntEnum
from functools import reduce

from rdkit.Chem import EState

from ._base import Descriptor
from ._util import parse_enum

try:
    import builtins
except ImportError:
    import __builtin__ as builtins


__all__ = ('AtomTypeEState',)


es_types = (
    'sLi', 'ssBe', 'ssssBe', 'ssBH', 'sssB', 'ssssB', 'sCH3', 'dCH2', 'ssCH2',
    'tCH', 'dsCH', 'aaCH', 'sssCH', 'ddC', 'tsC', 'dssC', 'aasC', 'aaaC',
    'ssssC', 'sNH3', 'sNH2', 'ssNH2', 'dNH', 'ssNH', 'aaNH', 'tN', 'sssNH',
    'dsN', 'aaN', 'sssN', 'ddsN', 'aasN', 'ssssN', 'sOH', 'dO', 'ssO', 'aaO',
    'sF', 'sSiH3', 'ssSiH2', 'sssSiH', 'ssssSi', 'sPH2', 'ssPH', 'sssP',
    'dsssP', 'sssssP', 'sSH', 'dS', 'ssS', 'aaS', 'dssS', 'ddssS', 'sCl',
    'sGeH3', 'ssGeH2', 'sssGeH', 'ssssGe', 'sAsH2', 'ssAsH', 'sssAs', 'sssdAs',
    'sssssAs', 'sSeH', 'dSe', 'ssSe', 'aaSe', 'dssSe', 'ddssSe', 'sBr', 'sSnH3',
    'ssSnH2', 'sssSnH', 'ssssSn', 'sI', 'sPbH3', 'ssPbH2', 'sssPbH', 'ssssPb'
)

es_type_set = set(es_types)


class EStateBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False


class EStateCache(EStateBase):
    __slots__ = ()

    def parameters(self):
        return ()

    def calculate(self):
        return EState.TypeAtoms(self.mol), EState.EStateIndices(self.mol)


class AggrType(IntEnum):
    __slots__ = ()

    count = 1
    sum = 2
    max = 3
    min = 4

    @property
    def as_argument(self):
        return self.name

aggr_names = (
    (AggrType.count, 'N'),
    (AggrType.sum, 'S'),
    (AggrType.max, 'MAX'),
    (AggrType.min, 'MIN'),
)


aggr_name_dict = dict(aggr_names)


class AtomTypeEState(EStateBase):
    r"""atom type e-state descriptor.

    :type type: str
    :param type: one of aggr_types

    :type estate: str
    :param estate: one of es_types

    :returns: NaN when type in ['min', 'max'] and :math:`N_{\rm X} = 0`

    References
        * :doi:`10.1021/ci00028a014`
    """
    __slots__ = ('_type', '_estate',)

    aggr_types = tuple(a.name for a in AggrType)

    es_types = es_types

    @classmethod
    def preset(cls):
        return (
            cls(a, t)
            for a in AggrType
            for t in es_types
        )

    def __str__(self):
        aggr = aggr_name_dict[self._type]

        return aggr + self._estate

    def parameters(self):
        return self._type, self._estate

    def __init__(self, type='count', estate='sLi'):
        assert estate in es_type_set

        self._type = parse_enum(AggrType, type)
        self._estate = estate

    def dependencies(self):
        return {'E': EStateCache()}

    def calculate(self, E):
        if self._type == AggrType.count:
            return reduce(lambda a, b: a + b, E[0]).count(self._estate)

        indices = map(
            lambda e: e[1],
            filter(lambda e: self._estate in e[0], zip(*E))
        )

        with self.rethrow_na(ValueError):
            return getattr(builtins, self._type.name)(indices)

    @property
    def rtype(self):
        r"""
        * 'count': :py:class:`int`
        * 'other': :py:class:`float`
        """

        return int if self._type == AggrType.count else float

    _extra_docs = 'aggr_types', 'es_types'



from ._base import Descriptor
from ._graph_matrix import DistanceMatrix as D
from ._matrix_attributes import methods, get_method

__all__ = ('DistanceMatrix',)


class DistanceMatrix(Descriptor):
    r"""distance matrix descriptor.

    :type type: str
    :param type: :ref:`matrix_aggregating_methods`
    """
    __slots__ = ('_type',)
    explicit_hydrogens = False

    @classmethod
    def preset(cls):
        return map(cls, methods)

    def __str__(self):
        return '{}_D'.format(self._type.__name__)

    def parameters(self):
        return self._type,

    def __init__(self, type='SpMax'):
        self._type = get_method(type)

    def dependencies(self):
        return {
            'result': self._type(
                D(self.explicit_hydrogens),
                self.explicit_hydrogens,
                self.kekulize,
            )
        }

    def calculate(self, result):
        return result

    rtype = float



import math
from itertools import chain

from rdkit import Chem

from ._base import Descriptor

__all__ = (
    'PathCount',
)


class PathCountBase(Descriptor):
    explicit_hydrogens = False


class PathCountCache(PathCountBase):
    __slots__ = ('_order', '_bonds')

    def parameters(self):
        return self._order,

    def __init__(self, order):
        self._order = order

    def _gen_bonds(self):
        self._bonds = [
            (b.GetBeginAtomIdx(), b.GetEndAtomIdx())
            for b in self.mol.GetBonds()
        ]

    def _bond_ids_to_atom_ids(self, p):
        it = iter(p)

        try:
            a0f, a0t = self._bonds[next(it)]
        except StopIteration:
            return []

        try:
            a1f, a1t = self._bonds[next(it)]
        except StopIteration:
            return a0f, a0t

        if a0f in [a1f, a1t]:
            path = [a0t, a0f]
            current = a1f if a0f == a1t else a1t
        else:
            path = [a0f, a0t]
            current = a1f if a0t == a1t else a1t

        for i in it:
            anf, ant = self._bonds[i]

            path.append(current)

            if anf == current:
                current = ant
            else:
                current = anf

        path.append(current)
        return path

    def calculate(self):
        l = 0
        pi = 0

        self._gen_bonds()

        for path in Chem.FindAllPathsOfLengthN(self.mol, self._order):
            aids = set()
            before = None
            w = 1

            for i in self._bond_ids_to_atom_ids(path):
                if i in aids:
                    break

                aids.add(i)

                if before is not None:
                    bond = self.mol.GetBondBetweenAtoms(before, i)
                    w *= bond.GetBondTypeAsDouble()

                before = i

            else:
                l += 1
                pi += w

        return l, pi


class PathCount(PathCountBase):
    r"""path count descriptor.

    :type order: int
    :param order: path order(number of bonds in path)

    :type pi: bool
    :param pi: calculate pi-path count

    :type total: bool
    :param total: total path count(1 to order)

    :type log: bool
    :param log: use log scale
    """

    @classmethod
    def preset(cls):
        return chain(
            (cls(o, False, False, False) for o in range(2, 11)),
            [cls(10, False, True, False)],
            (cls(o, True, False, True) for o in range(1, 11)),
            [cls(10, True, True, True)],
        )

    def __str__(self):
        base = 'T' if self._total else ''
        pi = 'piPC' if self._pi else 'MPC'

        return '{}{}{}'.format(base, pi, self._order)

    __slots__ = ('_order', '_pi', '_total', '_log',)

    def parameters(self):
        return self._order, self._pi, self._total, self._log

    def __init__(self, order=1, pi=False, total=False, log=False):
        assert order >= 0

        self._order = order
        self._pi = pi
        self._total = total
        self._log = log

    def dependencies(self):
        deps = {'PC': PathCountCache(self._order)}
        if self._total and self._order > 0:
            deps['acc'] = self.__class__(
                self._order - 1,
                self._pi,
                self._total,
            )

        return deps

    def calculate(self, PC, acc=None):
        if self._order == 0:
            return self.rtype(self.mol.GetNumAtoms())

        v = PC[1] if self._pi else PC[0]

        if acc is not None:
            v = acc + v

        if self._log:
            v = math.log(v + 1)

        return v

    @property
    def rtype(self):
        r"""
        * pi = True: :py:class:`float`
        * pi = False: :py:class:`int`
        """
        return float if self._pi else int



import numpy as np

from ._base import Descriptor
from ._atomic_property import AtomicProperty, get_properties

__all__ = ('ConstitutionalSum', 'ConstitutionalMean',)


class ConstitutionalSum(Descriptor):
    r"""sum of constitutional descriptor.

    .. math::
        S_p = \sum^A_{i=1} \frac{p_i}{p_{\rm C}}

    where
    :math:`p_i` is atomic property of i-th atom,
    :math:`p_{\rm C}` is atomic property of carbon

    :type prop: :py:class:`str` or :py:class:`function`
    :param prop: :ref:`atomic_properties`
    """
    __slots__ = ('_prop',)

    @classmethod
    def preset(cls):
        return map(cls, get_properties())

    def parameters(self):
        return self._prop,

    def __init__(self, prop='v'):
        self._prop = AtomicProperty(self.explicit_hydrogens, prop)

    _prefix = 'S'

    def __str__(self):
        return '{}{}'.format(self._prefix, self._prop.as_argument)

    def dependencies(self):
        return {'P': self._prop}

    def calculate(self, P):
        return np.sum(P / self._prop.carbon)

    rtype = float


class ConstitutionalMean(ConstitutionalSum):
    r"""mean of constitutional descriptor.

    .. math::
        M_p = \frac{S_p}{A}

    :type prop: :py:class:`str` or :py:class:`function`
    :param prop: :ref:`atomic_properties`

    :rtype: float
    """
    __slots__ = ('_prop',)
    _prefix = 'M'

    @classmethod
    def preset(cls):
        return map(cls, get_properties())

    def dependencies(self):
        return {'S': ConstitutionalSum(self._prop)}

    def calculate(self, S):
        return S / self.mol.GetNumAtoms()



from __future__ import division

import networkx as nx

from ._base import Descriptor
from .RingCount import Rings

__all__ = ('Framework',)


class FrameworkCache(Descriptor):
    __slots__ = ()

    def parameters(self):
        return ()

    def dependencies(self):
        return {'Rs': Rings()}

    def calculate(self, Rs):
        G = nx.Graph()
        Rd = {i: ('R', Ri) for Ri, R in enumerate(Rs) for i in R}
        R = list(set(Rd.values()))
        NR = len(R)

        for bond in self.mol.GetBonds():
            a = bond.GetBeginAtomIdx()
            b = bond.GetEndAtomIdx()

            a = Rd.get(a, ('A', a))
            b = Rd.get(b, ('A', b))

            G.add_edge(a, b)

        linkers = set()
        for Ri, Rj in ((i, j) for i in range(NR) for j in range(i + 1, NR)):
            Ra, Rb = R[Ri], R[Rj]
            try:
                linkers.update(i for t, i in nx.shortest_path(G, Ra, Rb) if t == 'A')
            except nx.NetworkXNoPath:
                pass

        return linkers, Rs


class Framework(Descriptor):
    r"""molecular framework ratio descriptor.

    .. math::

        f_{\rm MF} = \frac{N_{\rm MF}}{N}

    where
    :math:`N_{\rm MF}` is number of atoms in molecular framework,
    :math:`N` is number of all atoms.

    References
        * :doi:`10.1021/jm9602928`
    """
    __slots__ = ()

    @classmethod
    def preset(cls):
        yield cls()

    def __str__(self):
        return 'fMF'

    def parameters(self):
        return ()

    def dependencies(self):
        return {'F': FrameworkCache()}

    def calculate(self, F):
        linkers, rings = F
        Nmf = len(linkers) + len({i for ring in rings for i in ring})
        N = self.mol.GetNumAtoms()

        return Nmf / N

    rtype = float



import numpy as np

from ._base import Descriptor

__all__ = ('VertexAdjacencyInformation',)


class VertexAdjacencyInformation(Descriptor):
    r"""vertex adjacency information descriptor.

    .. math::
        {\rm VAdjMat} = 1 + \log_2(m)

    where :math:`m` is number of heavy-heavy bonds.

    :returns: :math:`m = 0`
    """
    __slots__ = ()

    @classmethod
    def preset(cls):
        yield cls()

    explicit_hydrogens = False

    def __str__(self):
        return 'VAdjMat'

    def parameters(self):
        return ()

    def calculate(self):
        m = sum(
            1
            for b in self.mol.GetBonds()
            if b.GetBeginAtom().GetAtomicNum() != 1 and
            b.GetEndAtom().GetAtomicNum() != 1
        )

        with self.rethrow_zerodiv():
            return 1 + np.log2(m)

    rtype = float



from __future__ import division

from .Chi import ChiCache
from ._base import Descriptor

__all__ = ('KappaShapeIndex1', 'KappaShapeIndex2', 'KappaShapeIndex3',)


class KappaShapeIndexBase(Descriptor):
    explicit_hydrogens = False

    @classmethod
    def preset(cls):
        yield cls()

    def __str__(self):
        return 'Kier{}'.format(self._order)

    def parameters(self):
        return ()

    def __init__(self):
        self._order = int(self.__class__.__name__[-1])

    def dependencies(self):
        return {'Chi': ChiCache(self._order)}

    def _common(self, Chi):
        P = len(Chi.path)

        A = self.mol.GetNumAtoms()
        Pmin = A - self._order

        return P, A, Pmin

    rtype = float


class KappaShapeIndex1(KappaShapeIndexBase):
    r"""Kappa shape index 1 descriptor.

    :returns: NaN when :math:`N_{\rm Chi-path} = 0`
    """

    __slots__ = ()

    def calculate(self, Chi):
        P, A, Pmin = self._common(Chi)
        Pmax = 0.5 * A * (A - 1)

        with self.rethrow_zerodiv():
            return 2 * Pmax * Pmin / (P * P)


class KappaShapeIndex2(KappaShapeIndexBase):
    r"""Kappa shape index 2 descriptor.

    :returns: NaN when :math:`N_{\rm Chi-path} = 0`
    """

    __slots__ = ()

    def calculate(self, Chi):
        P, A, Pmin = self._common(Chi)
        Pmax = 0.5 * (A - 1) * (A - 2)

        with self.rethrow_zerodiv():
            return 2 * Pmax * Pmin / (P * P)


class KappaShapeIndex3(KappaShapeIndexBase):
    r"""Kappa shape index 3 descriptor.

    :returns: NaN when :math:`N_{\rm Chi-path} = 0`
    """

    __slots__ = ()

    def calculate(self, Chi):
        P, A, Pmin = self._common(Chi)

        if A % 2 == 0:
            Pmax = 0.25 * (A - 2) ** 2
        else:
            Pmax = 0.25 * (A - 1) * (A - 3)

        with self.rethrow_zerodiv():
            return 4 * Pmax * Pmin / (P * P)



from rdkit import Chem
from numpy.testing import assert_almost_equal

from mordred import _atomic_property

_data = {
    '>C<': 'CC(C)(C)C',
    '>CH-': 'CC(C)C',
    '-CH2-': 'CCC',
    '=C<': 'C=C(C)C',
    '-CH3': 'CC',
    '=CH-': 'CC=C',
    '>N-': 'CN(C)C',
    '#C-': 'C#CC',
    '-NH-': 'CNC',
    '=CH2': 'C=C',
    '=N-': 'C=NC',
    '-O-': 'COC',
    '#CH': 'C#C',
    '-NH2': 'CN',
    '=NH': 'C=N',
    '#N': 'C#N',
    '-OH': 'CO',
    '=O': 'C=O',
    '-F': 'CF',
    '-SH': 'CS',
    '-S-': 'CSC',
    '=S': 'C=S',
    '-Cl': 'CCl',
    '-Br': 'CBr',
    '-I': 'CI',
}

explicitHs = {key: Chem.AddHs(Chem.MolFromSmiles(smi)).GetAtomWithIdx(1)
              for key, smi in _data.items()}

implicitHs = {key: Chem.MolFromSmiles(smi).GetAtomWithIdx(1)
              for key, smi in _data.items()}


def make_cases(results, fn, decimal=7):
    for key, result in results.items():
        yield assert_almost_equal, result, fn(explicitHs[key]), decimal, key
        yield assert_almost_equal, result, fn(implicitHs[key]), decimal, key


def test_sigma():
    results = {
        '>C<': 4,
        '>CH-': 3,
        '-CH2-': 2,
        '=C<': 3,
        '-CH3': 1,
        '=CH-': 2,
        '>N-': 3,
        '#C-': 2,
        '-NH-': 2,
        '=CH2': 1,
        '=N-': 2,
        '-O-': 2,
        '#CH': 1,
        '-NH2': 1,
        '=NH': 1,
        '#N': 1,
        '-OH': 1,
        '=O': 1,
        '-F': 1,
        '-SH': 1,
        '-S-': 2,
        '=S': 1,
        '-Cl': 1,
        '-Br': 1,
        '-I': 1,
    }
    for test in make_cases(results, _atomic_property.get_sigma_electrons):
        yield test


def test_valence_sigma():
    results = {
        '>C<': 4,
        '>CH-': 3,
        '-CH2-': 2,
        '=C<': 4,
        '-CH3': 1,
        '=CH-': 3,
        '>N-': 5,
        '#C-': 4,
        '-NH-': 4,
        '=CH2': 2,
        '=N-': 5,
        '-O-': 6,
        '#CH': 3,
        '-NH2': 3,
        '=NH': 4,
        '#N': 5,
        '-OH': 5,
        '=O': 6,
        '-F': 7,
        '-S-': 0.67,
        '-Cl': 0.78,
        '-Br': 0.26,
        '-I': 0.16,
    }
    for test in make_cases(results, _atomic_property.get_valence_electrons, 2):
        yield test


def test_IntrinsicState():
    results = {
        '>C<': 1.25,
        '>CH-': 1.3333,
        '-CH2-': 1.5,
        '=C<': 1.6666,
        '-CH3': 2.0,
        '=CH-': 2.0,
        '>N-': 2.0,
        '#C-': 2.5,
        '-NH-': 2.5,
        '=CH2': 3.0,
        '=N-': 3.0,
        '-O-': 3.5,
        '#CH': 4.0,
        '-NH2': 4.0,
        '=NH': 5.0,
        '#N': 6.0,
        '-OH': 6.0,
        '=O': 7.0,
        '-F': 8.0,
    }

    for test in make_cases(results, _atomic_property.get_intrinsic_state, 3):
        yield test



r"""surface area subpackage.

.. code:: console

    $ python -m mordred.surface_area -h
    usage: python -m mordred.surface_area [-h] [-s R] [-l L] FILE

    positional arguments:
      FILE                  input sdf/mol file

    optional arguments:
      -h, --help            show this help message and exit
      -s R, --solvent-radius R
                            solvent radius (default: 1.4)
      -l L, --mesh-level L  mesh level (default: 5)
"""

from ._sasa import SurfaceArea

__all__ = ('SurfaceArea',)



r"""Extended Topochemical Atom(ETA) descriptor.

References
    * :doi:`10.1021/ci0342066`
    * :doi:`10.1016/j.jhazmat.2013.03.023`
"""
from __future__ import division

import numpy as np
from rdkit import Chem

from . import _atomic_property as ap
from ._base import Descriptor
from ._util import atoms_to_numpy
from .RingCount import RingCount
from ._graph_matrix import DistanceMatrix

__all__ = (
    'EtaCoreCount', 'EtaShapeIndex',
    'EtaVEMCount',
    'EtaCompositeIndex', 'EtaFunctionalityIndex', 'EtaBranchingIndex',

    'EtaDeltaAlpha',
    'EtaEpsilon', 'EtaDeltaEpsilon',
    'EtaDeltaBeta',
    'EtaPsi', 'EtaDeltaPsi',
)


class AlterMolecule(Descriptor):
    __slots__ = ('explicit_hydrogens', '_saturated',)

    kekulize = True
    require_connected = True

    def parameters(self):
        return self.explicit_hydrogens, self._saturated

    def __init__(self, explicit_hydrogens, saturated=False):
        self._saturated = saturated
        self.explicit_hydrogens = explicit_hydrogens

    def __str__(self):
        b = 'Saturated' if self._saturated else 'Reference'
        H = 'H' if self.explicit_hydrogens else ''

        return '{}Mol{}'.format(b, H)

    def calculate(self):
        new = Chem.RWMol(Chem.Mol())
        ids = {}
        for a in self.mol.GetAtoms():
            if a.GetAtomicNum() == 1:
                continue

            if self._saturated:
                new_a = Chem.Atom(a.GetAtomicNum())
                new_a.SetFormalCharge(a.GetFormalCharge())

            else:
                new_a = Chem.Atom(6)

            ids[a.GetIdx()] = new.AddAtom(new_a)

        for bond in self.mol.GetBonds():
            ai = bond.GetBeginAtom()
            aj = bond.GetEndAtom()

            if not self._saturated and (ai.GetDegree() > 4 or aj.GetDegree() > 4):
                self.fail(ValueError('bond degree greater then 4'))

            i = ids.get(ai.GetIdx())
            j = ids.get(aj.GetIdx())

            if i is not None and j is not None:
                if self._saturated and (ai.GetAtomicNum() != 6 or aj.GetAtomicNum() != 6):
                    order = bond.GetBondType()
                else:
                    order = Chem.BondType.SINGLE

                new.AddBond(i, j, order)

        new = Chem.Mol(new)
        if Chem.SanitizeMol(new, catchErrors=True) != 0:
            typ = 'saturated' if self._saturated else 'referense'
            self.fail(ValueError('cannot sanitize {} mol'.format(typ)))

        if self.explicit_hydrogens:
            new = Chem.AddHs(new)

        Chem.Kekulize(new)

        return new


class EtaBase(Descriptor):
    __slots__ = ()

    explicit_hydrogens = False
    kekulize = True
    require_connected = True

    rtype = float


class EtaCoreCount(EtaBase):
    r"""ETA core count descriptor.

    .. math::

        \alpha = \sum_{i = 1}^A \frac{Z_i - Z_i^v}{Z_i^v} \cdot \frac{1}{PN_i - 1}

    where :math:`Z_i` and :math:`Z_i^v` are number of total and valence electons,
    :math:`PN` is periodic number.

    :type averaged: bool
    :param averaged: averaged by number of heavy count

    :type reference: bool
    :param reference: use reference alkane
        (same graph structure, but all atoms are carbon and all bonds are single bond)

    :returns: reference and valence of any atoms > 4
    """
    __slots__ = ('_averaged', '_reference',)

    @classmethod
    def preset(cls):
        return map(cls, [False, True])

    def __str__(self):
        suffix = '_R' if self._reference else ''
        ave = "A" if self._averaged else ''

        return '{}ETA_alpha{}'.format(ave, suffix)

    def parameters(self):
        return self._averaged, self._reference

    def __init__(self, averaged=False, reference=False):
        self._averaged = averaged
        self._reference = reference

    def dependencies(self):
        if self._reference:
            return {'rmol': AlterMolecule(self.explicit_hydrogens)}

    def calculate(self, rmol=None):
        mol = rmol if self._reference else self.mol

        v = sum(ap.get_core_count(a) for a in mol.GetAtoms())
        if self._averaged:
            v /= mol.GetNumAtoms()

        return v


class EtaShapeIndex(EtaBase):
    r"""ETA shape index descriptor.

    .. math::
        {\rm shape}_t = \frac{\alpha_t}{\alpha}

    where :math:`\alpha_t` is p(alpha value only atoms which bond to 1 heavy atom),
    y(3), or x(4).

    :type type: str
    :param type: one of shape_types
    """
    __slots__ = ('_type',)

    shape_types = ('p', 'y', 'x',)
    _type_to_degree = dict(p=1, y=3, x=4)

    @classmethod
    def preset(cls):
        return (cls(t) for t in cls.shape_types)

    def __str__(self):
        return 'ETA_shape_{}'.format(self._type)

    def parameters(self):
        return self._type,

    def __init__(self, type='p'):
        assert type in self.shape_types

        self._type = type

    def dependencies(self):
        return {'a': EtaCoreCount(False)}

    def calculate(self, a):
        d = self._type_to_degree[self._type]

        return sum(
            ap.get_core_count(a)
            for a in self.mol.GetAtoms()
            if a.GetDegree() == d
        ) / a


class EtaVEMCount(EtaBase):
    r"""ETA VEM(valence electron mobile) count descriptor.

    .. math::
        \beta^{\rm s} = \frac{1}{2} \sum^A_{i=1} \beta^{\rm s}_i

        \beta^{\rm s}_i = \sum^A_{j = 1} x_{ij}\sigma_{ij}

        x_{ij} = \begin{cases}
            0.5  & \left( \left| \epsilon_i - \epsilon_j \right| \leq 0.3 \right) \\
            0.75 & \left( \left| \epsilon_i - \epsilon_j \right| > 0.3 \right)
        \end{cases}

        \epsilon_i = - \alpha_i + 0.3 Z^{\rm v}

    where :math:`\sigma_{ij}` is sigma bond count between i and j.

    .. math::
        \beta^{\rm ns\delta} = \sum^A_{i = 1} \beta^{\rm ns\delta}_i

    where :math:`\beta^{\rm ns\delta}_i` is
    0.5 if i-th atom is making resonance with an aromatic ring.

    .. math::
        \beta^{\rm ns} = \frac{1}{2} \sum^A_{i=1} \beta^{\rm ns}_i

        \beta^{\rm ns}_i = \sum^A_{j = 1} y_{ij}\pi_{ij} + \beta^{\rm ns\delta}_i

        y_{ij} = \begin{cases}
            2.0 & \left( {\rm ij\ is\ aromatic\ bond} \right) \\
            1.5 & \left( \left| \epsilon_i - \epsilon_j \right| > 0.3 \right) \\
            1.0 & \left( \left| \epsilon_i - \epsilon_j \right| \leq 0.3 \right)
        \end{cases}

    where :math:`\pi_{ij}` is pi bond count between i and j.

    .. math::
        \beta = \beta^{\rm s} + \beta^{\rm ns}

    :type type: str
    :param type: one of beta_types

    :type averaged: bool
    :param averaged: averaged by heavy atom count
    """
    __slots__ = ('_type', '_averaged',)

    @classmethod
    def preset(cls):
        return (
            cls(b, a)
            for b in cls.beta_types
            for a in [False, True]
        )

    def __str__(self):
        typ = '_{}'.format(self._type) if self._type else ''
        ave = 'A' if self._averaged else ''

        return '{}ETA_beta{}'.format(ave, typ)

    beta_types = ('', 's', 'ns', 'ns_d')

    def parameters(self):
        return self._type, self._averaged

    def __init__(self, type='', averaged=False):
        assert type in self.beta_types
        self._type = type
        self._averaged = averaged

    def _get_beta_s(self, atom):
        return ap.get_eta_beta_sigma(atom) / 2.0

    def _get_beta_ns_d(self, atom):
        v = ap.get_eta_beta_delta(atom)
        return v

    def _get_beta_ns(self, atom):
        return ap.get_eta_beta_non_sigma(atom) / 2.0 + self._get_beta_ns_d(atom)

    def _get_beta_(self, atom):
        return self._get_beta_s(atom) + self._get_beta_ns(atom)

    def calculate(self):
        getter = getattr(self, '_get_beta_' + self._type)

        if getter:
            v = sum(
                getter(a)
                for a in self.mol.GetAtoms()
            )

        if self._averaged:
            v /= self.mol.GetNumAtoms()

        return v


class EtaCompositeIndex(EtaBase):
    r"""ETA composite index descriptor.

    .. math::
        \eta = \sum_{i < j} \left( \frac{\gamma_i \gamma_j}{r_{ij}^2} \right)^{0.5}

        \gamma_i = \frac{\alpha_i}{\beta_i}

    where :math:`r_{ij}` is graph distance.

    .. math::
        \eta^{\rm local} = \sum_{i < j, r_{ij} = 1} \left( \gamma_i \gamma_j \right)^{0.5}

    :type reference: bool
    :param reference: use reference alkane.

    :type local: bool
    :param local: use :math:`\eta^{\rm local}`

    :type averaged: bool
    :param averaged: averaged

    :returns: reference and valence of any atoms > 4
    """
    __slots__ = ('_reference', '_local', '_averaged',)

    @classmethod
    def preset(cls):
        ft = [False, True]
        return (cls(r, l, a) for r in ft for l in ft for a in ft)

    def __str__(self):
        suffixes = []

        if self._reference:
            suffixes.append('R')

        if self._local:
            suffixes.append('L')

        if len(suffixes) > 0:
            suffix = '_' + ''.join(suffixes)
        else:
            suffix = ''

        ave = 'A' if self._averaged else ''

        return '{}ETA_eta{}'.format(ave, suffix)

    def parameters(self):
        return self._reference, self._local, self._averaged

    def __init__(self, reference=False, local=False, averaged=False):
        self._reference = reference
        self._local = local
        self._averaged = averaged

    def dependencies(self):
        deps = {'D': DistanceMatrix(self.explicit_hydrogens)}

        if self._reference:
            deps['rmol'] = AlterMolecule(self.explicit_hydrogens)

        return deps

    def calculate(self, D, rmol=None):
        mol = rmol if self._reference else self.mol

        if self._local:
            def checker(r):
                return r == 1
        else:
            def checker(r):
                return r != 0

        gamma = atoms_to_numpy(ap.get_eta_gamma, mol)

        v = sum(
            sum(
                np.sqrt(gamma[i] * gamma[j] / r ** 2)
                for j, r in enumerate(row)
                if i < j and checker(r)
            ) for i, row in enumerate(D)
        )

        if self._averaged:
            v /= mol.GetNumAtoms()

        return v


class EtaFunctionalityIndex(EtaBase):
    r"""ETA functionality index descriptor.

    .. math::
        \eta^{\rm F} = \eta^{\rm R} - \eta

    where :math:`\eta^{\rm R}` is eta value for reference alkane.

    :type local: bool
    :param local: use local eta

    :type averaged: bool
    :param averaged: averaged
    """
    __slots__ = ('_local', '_averaged',)

    @classmethod
    def preset(cls):
        return (
            cls(l, a)
            for l in [False, True]
            for a in [False, True]
        )

    def __str__(self):
        loc = 'L' if self._local else ''
        ave = 'A' if self._averaged else ''

        return '{}ETA_eta_F{}'.format(ave, loc)

    def parameters(self):
        return self._local, self._averaged

    def __init__(self, local=False, averaged=False):
        self._local = local
        self._averaged = averaged

    def dependencies(self):
        return {
            'eta': EtaCompositeIndex(local=self._local),
            'eta_R': EtaCompositeIndex(local=self._local, reference=True),
        }

    def calculate(self, eta, eta_R):
        v = eta_R - eta
        if self._averaged:
            v /= self.mol.GetNumAtoms()

        return v


class EtaBranchingIndex(EtaBase):
    r"""ETA branching index descriptor.

    .. math::
        \eta^{\rm B} = \eta^{\rm local,N} - \eta^{local,R} + 0.086 N^{\rm R}

    where :math:`\eta^{\rm local,N}` is :math:`\eta^{\rm local}` for normal alkane.
    :math:`N^{\rm R}` is ring count.

    :type ring: bool
    :param ring: use ring count or not

    :type averaged: bool
    :param averaged: averaged

    :returns: NaN when A < 2
    """
    __slots__ = ('_ring', '_averaged',)

    @classmethod
    def preset(cls):
        return (
            cls(r, a)
            for r in [False, True]
            for a in [False, True]
        )

    def __str__(self):
        ring = 'R' if self._ring else ''
        ave = 'A' if self._averaged else ''

        return '{}ETA_eta_B{}'.format(ave, ring)

    def parameters(self):
        return self._ring, self._averaged

    def __init__(self, ring=True, averaged=False):
        self._ring = ring
        self._averaged = averaged

    def dependencies(self):
        return {
            'eta_RL': EtaCompositeIndex(reference=True, local=True),
            'NR': RingCount() if self._ring else None,
        }

    def calculate(self, eta_RL, NR):
        N = self.mol.GetNumAtoms()

        if N <= 1:
            self.fail(ValueError('single atom'))
        elif N == 2:
            eta_NL = 1.0
        else:
            eta_NL = np.sqrt(2) + 0.5 * (N - 3)

        v = eta_NL - eta_RL + 0.086 * (NR or 0)

        if self._averaged:
            v /= N

        return v


class EtaDeltaAlpha(EtaBase):
    r"""ETA delta alpha descriptor.

    .. math::
        \Delta\alpha_{\rm A} = \max\left(\frac{\alpha - \alpha^{\rm R}}{A}, 0\right)

        \Delta\alpha_{\rm B} = \max\left(\frac{\alpha^{\rm R} - \alpha}{A}, 0\right)

    :type type: str
    :param type: one of delta_types
    """
    __slots__ = ('_type',)

    delta_types = ('A', 'B',)

    @classmethod
    def preset(cls):
        return (cls(t) for t in cls.delta_types)

    def __str__(self):
        return 'ETA_dAlpha_{}'.format(self._type)

    def parameters(self):
        return self._type,

    def __init__(self, type='A'):
        assert type in self.delta_types

        self._type = type

    def dependencies(self):
        return {
            'alpha': EtaCoreCount(),
            'alpha_R': EtaCoreCount(reference=True),
        }

    def calculate(self, alpha, alpha_R):
        if self._type == 'A':
            d = alpha - alpha_R
        else:
            d = alpha_R - alpha

        return max(d / self.mol.GetNumAtoms(), 0.0)


class EtaEpsilon(EtaBase):
    r"""ETA epsilon descriptor.

    .. math::
        \epsilon^i = \frac{\epsilon^i}{N^i} (i \leq 4)
        \epsilon^5 = \frac{\epsilon^2 + \epsilon^{\rm XH}}{N^2 + N^{\rm XH}}

    types(i)
        1
            all atoms
        2
            heavy atoms
        3
            all atoms of reference alkane
        4
            all atoms of saturated carbon skeleton(reduce C-C bonds)
        XH
            hydrogens bond to hetero atoms

    :type type: str
    :param type: one of epsilon_types

    :returns: type = 3 and valence of any atoms > 4
    """
    __slots__ = ('_type',)

    @classmethod
    def preset(cls):
        return map(cls, cls.epsilon_types)

    def __str__(self):
        return 'ETA_epsilon_{}'.format(self._type)

    @property
    def explicit_hydrogens(self):
        return self._type != 2

    epsilon_types = tuple(range(1, 6))

    def parameters(self):
        return self._type,

    def __init__(self, type=1):
        self._type = type

    def dependencies(self):
        if self._type == 3:
            return {'rmol': AlterMolecule(self.explicit_hydrogens)}
        elif self._type == 4:
            return {'rmol': AlterMolecule(self.explicit_hydrogens, True)}

    def calculate(self, rmol=None):
        mol = rmol if self._type in [3, 4] else self.mol

        if self._type == 5:
            eps = [
                ap.get_eta_epsilon(a)
                for a in mol.GetAtoms()
                if a.GetAtomicNum() != 1 or a.GetNeighbors()[0].GetAtomicNum() != 6
            ]
            return sum(eps) / len(eps)

        return sum(ap.get_eta_epsilon(a) for a in mol.GetAtoms()) / mol.GetNumAtoms()


class EtaDeltaEpsilon(EtaBase):
    r"""ETA delta epsilon descriptor.

    .. math::
        \Delta \epsilon^{\rm A} = \epsilon^1 - \epsilon^3

        \Delta \epsilon^{\rm B} = \epsilon^1 - \epsilon^4

        \Delta \epsilon^{\rm C} = \epsilon^3 - \epsilon^4

        \Delta \epsilon^{\rm D} = \epsilon^2 - \epsilon^5

    :type type: str
    :param type: one of delta_epsilon_types
    """
    __slots__ = ('_type',)

    @classmethod
    def preset(cls):
        return map(cls, cls.delta_epsilon_types)

    def __str__(self):
        return 'ETA_dEpsilon_{}'.format(self._type)

    delta_epsilon_types = tuple('ABCD')

    def parameters(self):
        return self._type,

    def __init__(self, type='A'):
        self._type = type

    _deps = dict(
        A=(1, 3),
        B=(1, 4),
        C=(3, 4),
        D=(2, 5),
    )

    def dependencies(self):
        L, R = self._deps[self._type]
        return {
            'L': EtaEpsilon(L),
            'R': EtaEpsilon(R),
        }

    def calculate(self, L, R):
        return L - R


class EtaDeltaBeta(EtaBase):
    r"""ETA delta beta descriptor.

    .. math::
        \Delta\beta = \beta^{\rm ns} - \beta^{\rm s}

    :type averaged: bool
    :param averaged: averaged
    """
    __slots__ = ('_averaged',)

    @classmethod
    def preset(cls):
        return (cls(a) for a in [False, True])

    def __str__(self):
        ave = 'A' if self._averaged else ''

        return '{}ETA_dBeta'.format(ave)

    def parameters(self):
        return self._averaged,

    def __init__(self, averaged=False):
        self._averaged = averaged

    def dependencies(self):
        return {
            'ns': EtaVEMCount('ns'),
            's': EtaVEMCount('s'),
        }

    def calculate(self, ns, s):
        v = ns - s

        if self._averaged:
            v /= self.mol.GetNumAtoms()

        return v


class EtaPsi(EtaBase):
    r"""ETA psi descriptor.

    .. math::
        \psi_1 = \frac{\alpha}{A \cdot \epsilon^2}
    """

    @classmethod
    def preset(cls):
        yield cls()

    def __str__(self):
        return 'ETA_psi_1'

    def parameters(self):
        return ()

    def dependencies(self):
        return {
            'a': EtaCoreCount(),
            'e': EtaEpsilon(2),
        }

    def calculate(self, a, e):
        return a / (self.mol.GetNumAtoms() * e)


class EtaDeltaPsi(EtaBase):
    r"""ETA delta psi descriptor.

    .. math::
        \Delta\psi_{\rm A} = \max\left(0.714 - \psi_1, 0\right)

        \Delta\psi_{\rm B} = \max\left(\psi_1 - 0.714, 0\right)

    :type type: str
    :param type: one of delta_psi_types
    """
    __slots__ = ('_type',)

    @classmethod
    def preset(cls):
        return map(cls, cls.delta_psi_types)

    def __str__(self):
        return 'ETA_dPsi_{}'.format(self._type)

    delta_psi_types = ('A', 'B',)

    def parameters(self):
        return self._type,

    def __init__(self, type='A'):
        assert type in self.delta_psi_types

        self._type = type

    def dependencies(self):
        return {'psi': EtaPsi()}

    def calculate(self, psi):
        L = 0.714
        R = psi

        if self._type == 'B':
            L, R = R, L

        return max(L - R, 0.0)



from multiprocessing import Pool

import six

from .._util import Capture
from .context import Context

calculator = None


def initializer(self):
    global calculator

    calculator = self


def worker(cxt):
    with Capture() as capture:
        r = list(calculator._calculate(cxt))

        return r, capture.result


def parallel(self, mols, nproc, nmols, quiet, ipynb, id):
    if six.PY3:
        def get_result(r):
            return r.get()
    else:
        # timeout: avoid python2 KeyboardInterrupt bug.
        # http://stackoverflow.com/a/1408476
        def get_result(r):
            return r.get(1e9)

    # without with-statement for compat. python2
    pool = Pool(nproc, initializer=initializer, initargs=(self,))

    def do_task(mol):
        args = Context.from_calculator(self, mol, id)
        return pool.apply_async(worker, (args,))

    try:
        with self._progress(quiet, nmols, ipynb) as bar:
            for result in [do_task(m) for m in mols]:
                r, err = get_result(result)

                for e in err:
                    e = e.rstrip()
                    if not e:
                        continue

                    bar.write(e)

                yield r
                bar.update()

    finally:
        pool.terminate()
        pool.join()



from rdkit.Chem.rdMolDescriptors import CalcNumRotatableBonds

from ._base import Descriptor
from .BondCount import BondCount

__all__ = (
    'RotatableBondsCount', 'RotatableBondsRatio',
)


class RotatableBondsBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False

    @classmethod
    def preset(cls):
        yield cls()

    def parameters(self):
        return ()


class RotatableBondsCount(RotatableBondsBase):
    r"""ratatable bonds count descriptor(rdkit wrapper)."""
    __slots__ = ()

    def __str__(self):
        return 'nRot'

    def calculate(self):
        return CalcNumRotatableBonds(self.mol)

    rtype = int


class RotatableBondsRatio(RotatableBondsBase):
    r"""rotatable bonds ratio descriptor.

    .. math::
        {\rm RotRatio} = \frac{N_{\rm rotatable bonds}}{N_{\rm bonds}}

    :returns: NaN when :math:`N_{\rm bonds} = 0`
    """
    __slots__ = ()

    def __str__(self):
        return 'RotRatio'

    def dependencies(self):
        return {
            'nRot': RotatableBondsCount(),
            'nB': BondCount('heavy'),
        }

    def calculate(self, nRot, nB):
        with self.rethrow_zerodiv():
            return float(nRot) / float(nB)

    rtype = float



from collections import namedtuple

import numpy as np
from six import string_types

from ._base import Descriptor

Eig = namedtuple('eigen', 'val vec min max')

methods = []


def method(cls):
    methods.append(cls)
    cls.as_argument = cls.__name__
    return cls


class Common(Descriptor):
    __slots__ = 'matrix', 'explicit_hydrogens', 'kekulize'
    require_connected = True

    def parameters(self):
        return self.matrix, self.explicit_hydrogens, self.kekulize

    def __init__(self, matrix, explicit_hydrogens, kekulize):
        self.matrix = matrix
        self.explicit_hydrogens = explicit_hydrogens
        self.kekulize = kekulize

    @property
    def _key_args(self):
        return (
            self.matrix,
            self.explicit_hydrogens,
            self.kekulize
        )

    def dependencies(self):
        return {'eig': Eigen(*self._key_args)}

    @property
    def _eig(self):
        return Eigen(*self._key_args)

    @property
    def _SpMax(self):
        return SpMax(*self._key_args)

    @property
    def _SpMean(self):
        return SpMean(*self._key_args)

    @property
    def _SpAD(self):
        return SpAD(*self._key_args)

    @property
    def _VE1(self):
        return VE1(*self._key_args)

    @property
    def _VR1(self):
        return VR1(*self._key_args)

    def __str__(self):
        n = self.__class__.__name__

        if self.kekulize:
            n += 'K'

        if self.explicit_hydrogens:
            n += 'H'

        return n


class Eigen(Common):
    __slots__ = ()

    def dependencies(self):
        return {'matrix': self.matrix}

    def calculate(self, matrix):
        if matrix is None:
            raise ValueError('matrix is None')

        w, v = np.linalg.eig(matrix)

        if np.iscomplexobj(w):
            w = w.real

        if np.iscomplexobj(v):
            v = v.real

        i_min = np.argmin(w)
        i_max = np.argmax(w)

        return Eig(w, v, i_min, i_max)


@method
class SpAbs(Common):
    __slots__ = ()

    def calculate(self, eig):
        return np.abs(eig.val).sum()


@method
class SpMax(Common):
    __slots__ = ()

    def calculate(self, eig):
        return eig.val[eig.max]


@method
class SpDiam(Common):
    __slots__ = ()

    def dependencies(self):
        return {
            'eig': self._eig,
            'SpMax': self._SpMax,
        }

    def calculate(self, SpMax, eig):
        return SpMax - eig.val[eig.min]


class SpMean(Common):
    __slots__ = ()

    def calculate(self, eig):
        return np.mean(eig.val)


@method
class SpAD(Common):
    __slots__ = ()

    def dependencies(self):
        return {
            'eig': self._eig,
            'SpMean': self._SpMean,
        }

    def calculate(self, eig, SpMean):
        return np.abs(eig.val - SpMean).sum()


@method
class SpMAD(Common):
    __slots__ = ()

    def dependencies(self):
        return {'SpAD': self._SpAD}

    def calculate(self, SpAD):
        return SpAD / self.mol.GetNumAtoms()


@method
class LogEE(Common):
    __slots__ = ()

    def calculate(self, eig):
        # log sum exp: https://hips.seas.harvard.edu/blog/2013/01/09/computing-log-sum-exp
        a = np.maximum(eig.val[eig.max], 0)
        sx = np.exp(eig.val - a).sum() + np.exp(-a)
        return a + np.log(sx)


@method
class SM1(Common):
    __slots__ = ()

    def calculate(self, eig):
        return eig.val.sum()


@method
class VE1(Common):
    __slots__ = ()

    def calculate(self, eig):
        return np.abs(eig.vec[:, eig.max]).sum()


@method
class VE2(Common):
    __slots__ = ()

    def dependencies(self):
        return {'VE1': self._VE1}

    def calculate(self, VE1):
        return VE1 / self.mol.GetNumAtoms()


@method
class VE3(Common):
    __slots__ = ()

    def dependencies(self):
        return {'VE1': self._VE1}

    def calculate(self, VE1):
        with self.rethrow_zerodiv():
            return np.log(0.1 * self.mol.GetNumAtoms() * VE1)


@method
class VR1(Common):
    __slots__ = ()

    def calculate(self, eig):
        s = 0.0

        for bond in self.mol.GetBonds():
            i = bond.GetBeginAtomIdx()
            j = bond.GetEndAtomIdx()

            s += (eig.vec[i, eig.max] * eig.vec[j, eig.max]) ** -0.5

        return s


@method
class VR2(Common):
    __slots__ = ()

    def dependencies(self):
        return {'VR1': self._VR1}

    def calculate(self, VR1):
        return VR1 / self.mol.GetNumAtoms()


@method
class VR3(Common):
    __slots__ = ()

    def dependencies(self):
        return {'VR1': self._VR1}

    def calculate(self, VR1):
        with self.rethrow_zerodiv():
            return np.log(0.1 * self.mol.GetNumAtoms() * VR1)


method_dict = {m.__name__: m for m in methods}


def get_method(n):
    if isinstance(n, string_types):
        return method_dict[n]

    return n



import numpy as np

from ._base import Descriptor
from ._util import atoms_to_numpy

__all__ = ('MomentOfInertia',)


class MomentOfInertiaBase(Descriptor):
    __slots__ = ()
    require_3D = True

    def _numpy(self):
        ws = atoms_to_numpy(lambda a: a.GetMass(), self.mol)
        ps = self.coord - np.sum(ws[:, np.newaxis] * self.coord, axis=0) / np.sum(ws)

        return ws, ps


class PrincipalAxis(MomentOfInertiaBase):
    __slots__ = ()

    def parameters(self):
        return ()

    def calculate(self):
        ws, ps = self._numpy()

        I = np.sum(
            -ws[:, np.newaxis, np.newaxis] *
            (ps[:, np.newaxis] * ps[:, :, np.newaxis]),
            axis=0
        )

        diag = np.sum(
            ws[:, np.newaxis] *
            (np.sum(ps ** 2, axis=1)[:, np.newaxis] - ps ** 2),
            axis=0
        )

        np.fill_diagonal(I, diag)
        return np.sort(np.linalg.eig(I)[0])[::-1]


class MomentOfInertia(MomentOfInertiaBase):
    __slots__ = '_axis',

    @classmethod
    def preset(cls):
        return map(cls, cls.axes)

    def __str__(self):
        return 'MOMI-{}'.format(self._axis)

    def parameters(self):
        return self._axis,

    axes = ('X', 'Y', 'Z')
    _axis_to_index = {a: i for i, a in enumerate(axes)}

    def __init__(self, axis='X'):
        assert axis in self.axes
        self._axis = axis

    def dependencies(self):
        return {'I': PrincipalAxis()}

    def calculate(self, I):
        return I[self._axis_to_index[self._axis]]

    rtype = float



import numpy as np
import six
from abc import ABCMeta, abstractproperty


class MissingValueBase(six.with_metaclass(ABCMeta, object)):
    '''Base class of missing values.

    Args:
        error (Exception): error object
        stack (callstack)
    '''
    __slots__ = 'error', 'stack'

    def __reduce_ex__(self, version):
        return self.__class__, (self.error, self.stack)

    def __init__(self, error, stack):
        self.error = error
        self.stack = stack

    def __float__(self):
        return np.nan

    def __add__(self, other):
        return np.nan

    def __sub__(self, other):
        return np.nan

    def __str__(self):
        return '{} ({})'.format(self.error, '/'.join(str(d) for d in self.stack))

    @abstractproperty
    def header(self):
        '''header of warning message

        Returns:
            str
        '''
        raise NotImplementedError('require header')


class Missing(MissingValueBase):
    '''known errored value'''
    __slots__ = ()
    header = 'Missing'


class Error(MissingValueBase):
    '''unknown errored value'''
    __slots__ = ()
    header = 'ERROR'


class MordredException(Exception):
    __slots__ = ()


class MultipleFragments(MordredException):
    '''multiple fragments detected on require_connected Descriptor.'''
    __slots__ = ()

    def __str__(self):
        return 'multiple fragments'


class Missing3DCoordinate(MordredException):
    '''missing 3D coordinate on require_3D Descriptor.'''
    __slots__ = ()

    def __str__(self):
        return 'missing 3D coordinate'


class DuplicatedDescriptorName(MordredException):
    '''duplicated string replisantation of descriptor'''

    __slots__ = ()

    def __init__(self, a, b):
        self.a = a
        self.b = b

    def __str__(self):
        return 'duplicated descriptor name: {!r} and {!r}'.format(self.a, self.b)



import os
import warnings

from importlib import import_module
from ..error import MissingValueBase

from .descriptor import Descriptor
from .calculator import Calculator, get_descriptors_from_module
from .parallel import parallel


__all__ = (
    'all_descriptors',
    'Descriptor',
    'Calculator',
    'get_descriptors_from_module',
)


def all_descriptors():
    r"""**[deprecated]** use mordred.descriptors module instead.

    yield all descriptor modules.

    :returns: all modules
    :rtype: :py:class:`Iterator` (:py:class:`Descriptor`)
    """
    warnings.warn('all_descriptors() is deprecated, use mordred.descriptors module instead', DeprecationWarning, stacklevel=2)
    base_dir = os.path.dirname(os.path.dirname(__file__))

    for name in os.listdir(base_dir):
        name, ext = os.path.splitext(name)
        if name[:1] == '_' or ext != '.py' or name == 'descriptors':
            continue

        yield import_module('..' + name, __package__)


def _Descriptor__call__(self, mol, id=-1):
    r"""calculate single descriptor value.
    :type id: int
    :param id: conformer id

    :returns: descriptor result
    :rtype: scalar
    """
    v = Calculator(self)(mol, id)[0]
    if isinstance(v, MissingValueBase):
        raise v.error

    return v


Descriptor.__call__ = _Descriptor__call__
Calculator._parallel = parallel



import pickle

from nose.tools import eq_
from rdkit.Chem import AllChem as Chem
from numpy.testing import assert_almost_equal

from mordred import Calculator, descriptors
from mordred.error import MissingValueBase


def test_parallel():
    calc = Calculator(descriptors)
    mols = list(map(Chem.AddHs, [
        Chem.MolFromSmiles('c1ccccc1'),
        Chem.MolFromSmiles('C1=CC(=C(C=C1C2=C(C=C3C(=CC(=CC3=[O+]2)O)O)O)O)O'),
        Chem.MolFromSmiles('CCCCCC'),
    ]))

    for mol in mols:
        Chem.EmbedMolecule(mol)

    for serial, parallel in zip(calc.map(mols, nproc=1, quiet=True),
                                calc.map(mols, quiet=True)):
        for d, s, p in zip(calc.descriptors, serial, parallel):
            if isinstance(s, MissingValueBase):
                yield eq_, pickle.dumps(s), pickle.dumps(p), str(d)
            else:
                yield assert_almost_equal, s, p, 7, str(d)



from rdkit.Chem.Descriptors import ExactMolWt

from ._base import Descriptor

__all__ = (
    'Weight',
)


class Weight(Descriptor):
    r"""molecular weight descriptor.

    :type averaged: bool
    :param averaged: averaged by number of atom
    """
    __slots__ = ('_averaged',)
    explicit_hydrogens = True

    @classmethod
    def preset(cls):
        yield cls(False)
        yield cls(True)

    def __str__(self):
        return 'AMW' if self._averaged else 'MW'

    def parameters(self):
        return self._averaged,

    def __init__(self, averaged=False):
        self._averaged = averaged

    def calculate(self):
        w = ExactMolWt(self.mol)
        if self._averaged:
            w /= self.mol.GetNumAtoms()

        return w

    rtype = float



from rdkit import Chem
from numpy.testing import assert_almost_equal

from mordred import Calculator, ExtendedTopochemicalAtom

references = {
    # Roy et. al. Quantitative Structure-Activity Relationships in Drug Design, Predictive Toxicology, and Risk Assessment. p. 65
    'O=Cc1c(Cl)cccc1': {
        "ETA_alpha": 4.547,
        "AETA_alpha": 0.505,
        "ETA_shape_p": 0.230,
        "ETA_shape_y": 0.220,
        "ETA_shape_x": 0.000,
        "ETA_dAlpha_A": 0.005,
        "ETA_dAlpha_B": 0.000,
        "ETA_eta": 3.536,
        "ETA_eta_R": 9.998,
        "AETA_eta_F": 0.718,
        "ETA_eta_L": 1.543,
        "ETA_eta_RL": 4.343,
        "ETA_eta_FL": 2.800,
        "AETA_eta_BR": 0.018,
        "ETA_epsilon_1": 0.661,
        "ETA_epsilon_2": 0.861,
        "ETA_epsilon_3": 0.433,
        "ETA_epsilon_4": 0.553,
        "ETA_epsilon_5": 0.861,
        "ETA_dEpsilon_A": 0.228,
        "ETA_dEpsilon_B": 0.108,
        "ETA_dEpsilon_C": -0.119,
        "ETA_dEpsilon_D": 0.000,
        "AETA_beta_s": 0.556,
        "AETA_beta_ns": 0.889,
        "AETA_beta": 1.444,
        "AETA_dBeta": 0.333,
        "AETA_beta_ns_d": 0.056,
        "ETA_psi_1": 0.586,
        "ETA_dPsi_A": 0.128,
        "ETA_dPsi_B": 0.000,
    },
    'n1cc(O)ccc1': {
        "ETA_alpha": 3.233,
        "AETA_alpha": 0.462,
        "ETA_shape_p": 0.103,
        "ETA_shape_y": 0.155,
        "ETA_shape_x": 0.000,
        "ETA_dAlpha_A": 0.000,
        "ETA_dAlpha_B": 0.038,
        "ETA_eta": 2.048,
        "ETA_eta_R": 6.626,  # original:6.267. use sum of [eta_R]_i
        "AETA_eta_F": 0.654,
        "ETA_eta_L": 1.073,
        "ETA_eta_RL": 3.394,
        "ETA_eta_FL": 2.321,
        "AETA_eta_BR": 0.015,
        "ETA_epsilon_1": 0.631,
        "ETA_epsilon_2": 0.867,
        "ETA_epsilon_3": 0.433,
        "ETA_epsilon_4": 0.548,
        "ETA_epsilon_5": 0.796,
        "ETA_dEpsilon_A": 0.197,
        "ETA_dEpsilon_B": 0.083,
        "ETA_dEpsilon_C": -0.115,
        "ETA_dEpsilon_D": 0.071,
        "AETA_beta_s": 0.607,
        "AETA_beta_ns": 0.929,
        "AETA_beta": 1.536,
        "AETA_dBeta": 0.321,
        "AETA_beta_ns_d": 0.071,
        "ETA_psi_1": 0.533,
        "ETA_dPsi_A": 0.181,
        "ETA_dPsi_B": 0.000,
    },
# Roy, et. al. Understanding the Basics of QSAR for Applications in Pharmaceutical Sciences and Risk Assessment. p.145
    'O=Cc1ccccc1': {
        "AETA_alpha": 0.479,
        "ETA_shape_p": 0.087,
        "ETA_shape_y": 0.130,
        "ETA_shape_x": 0.000,
        "ETA_dAlpha_A": 0.000,
        "ETA_dAlpha_B": 0.021,
        "ETA_eta": 2.576,
        "ETA_eta_R": 8.023,
        "AETA_eta_F": 0.681,
        "ETA_eta_L": 1.303,
        "AETA_eta_BR": 0.009,
        "ETA_epsilon_1": 0.583,
        "ETA_epsilon_2": 0.796,
        "ETA_epsilon_3": 0.433,
        "ETA_epsilon_4": 0.498,
        "ETA_epsilon_5": 0.796,
        "ETA_dEpsilon_A": 0.150,
        "ETA_dEpsilon_B": 0.085,
        "ETA_dEpsilon_C": -0.065,
        "ETA_dEpsilon_D": 0.000,
        "AETA_beta_s": 0.531,
        "AETA_beta_ns": 0.938,
        "AETA_beta": 1.469,
        "AETA_dBeta": 0.406,
        "AETA_beta_ns_d": 0.000,
        "ETA_psi_1": 0.602,
        "ETA_dPsi_A": 0.112,
        "ETA_dPsi_B": 0.000,
    },
}

def test_ETA():
    calc = Calculator(ExtendedTopochemicalAtom)

    for smi, desireds in references.items():
        mol = Chem.MolFromSmiles(smi)
        actuals = {str(d): v for d, v in zip(calc.descriptors, calc(mol))}

        for name, desired in desireds.items():
            yield assert_almost_equal, actuals[name], desired, 2, '{} of {}'.format(name, smi)



from itertools import chain, groupby

import numpy as np

from ._base import Descriptor
from ._graph_matrix import DistanceMatrix

__all__ = (
    'InformationContent',
    'TotalIC', 'StructuralIC', 'BondingIC', 'ComplementaryIC',
    'ModifiedIC', 'ZModifiedIC',
)


class BFSTree(object):
    __slots__ = ('tree', 'visited', 'bonds', 'atoms')

    def __init__(self, mol):
        self.tree = {}
        self.visited = set()

        self.bonds = {}
        for b in mol.GetBonds():
            s = b.GetBeginAtomIdx()
            d = b.GetEndAtomIdx()
            t = b.GetBondType()

            self.bonds[s, d] = t
            self.bonds[d, s] = t

        self.atoms = [(a.GetAtomicNum(), a.GetDegree(), a.GetNeighbors()) for a in mol.GetAtoms()]

    def reset(self, i):
        self.tree.clear()
        self.visited.clear()

        self.tree[i] = ()
        self.visited.add(i)

    def expand(self):
        self._expand(self.tree)

    def _expand(self, tree):
        for src, dst in list(tree.items()):
            self.visited.add(src)

            if dst is ():
                tree[src] = {
                    n.GetIdx(): ()
                    for n in self.atoms[src][2]
                    if n.GetIdx() not in self.visited
                }

            else:
                self._expand(dst)

    def _code(self, tree, before, trail):
        if len(tree) == 0:
            yield trail

        else:
            for src, dst in tree.items():

                code = []
                if before is not None:
                    bt = self.bonds[before, src]
                    code.append(bt)

                code.append(self.atoms[src][:2])

                nxt = tuple(chain(trail, code))
                for t in self._code(dst, src, nxt):
                    yield t

    def get_code(self, i, order):
        self.reset(i)

        for _ in range(order):
            self.expand()

        return tuple(sorted(self._code(self.tree, None, ())))


class InformationContentBase(Descriptor):
    __slots__ = ('_order',)
    kekulize = True

    def __str__(self):
        return self._name + str(self._order)

    @classmethod
    def preset(cls):
        return (cls(o) for o in range(6))

    def parameters(self):
        return self._order,

    def __init__(self, order=0):
        self._order = order

    rtype = float


class Ag(InformationContentBase):
    __slots__ = ('_order',)

    @classmethod
    def preset(cls):
        return ()

    _name = 'Ag'

    def dependencies(self):
        return {'D': DistanceMatrix(self.explicit_hydrogens)}

    def calculate(self, D):
        if self._order == 0:
            atoms = [a.GetAtomicNum() for a in self.mol.GetAtoms()]
        else:
            tree = BFSTree(self.mol)
            atoms = [
                tree.get_code(i, self._order)
                for i in range(self.mol.GetNumAtoms())
            ]

        ad = {a: i for i, a in enumerate(atoms)}
        Ags = [(k, sum(1 for _ in g)) for k, g in groupby(sorted(atoms))]
        Nags = len(Ags)
        return (
            np.fromiter((ad[k] for k, _ in Ags), 'int', Nags),
            np.fromiter((ag for _, ag in Ags), 'float', Nags),
        )

    rtype = None


def _shannon_entropy_term(a):
    return a * np.log2(a)


shannon_entropy_term = np.vectorize(_shannon_entropy_term)


def shannon_entropy(a, w=1):
    N = np.sum(a)
    return -np.sum(w * shannon_entropy_term(a / N))


class InformationContent(InformationContentBase):
    r"""neighborhood information content descriptor.

    :type order: int
    :param order: order(number of edge) of subgraph
    """
    __slots__ = ()

    _name = 'IC'

    def dependencies(self):
        return {'iAgs': Ag(self._order)}

    def calculate(self, iAgs):
        _, Ags = iAgs
        return shannon_entropy(Ags)


class TotalIC(InformationContentBase):
    r"""neighborhood total information content descriptor.

    .. math::
        {\rm TIC}_m = A \cdot {\rm IC}_m

    :type order: int
    :param order: order(number of edge) of subgraph
    """
    __slots__ = ()

    _name = 'TIC'

    def dependencies(self):
        return {'ICm': InformationContent(self._order)}

    def calculate(self, ICm):
        A = self.mol.GetNumAtoms()

        return A * ICm


class StructuralIC(TotalIC):
    r"""structural information content descriptor.

    .. math::
        {\rm SIC}_m = \frac{{\rm IC}_m}{\log_2 A}

    :type order: int
    :param order: order(number of edge) of subgraph
    """
    __slots__ = ()

    _name = 'SIC'

    def calculate(self, ICm):
        d = np.log2(self.mol.GetNumAtoms())

        with self.rethrow_zerodiv():
            return ICm / d


class BondingIC(TotalIC):
    r"""bonding information content descriptor.

    .. math::
        {\rm BIC}_m = \frac{{\rm IC}_m}{\log_2 \sum^B_{b=1} \pi^{*}_b}

    :type order: int
    :param order: order(number of edge) of subgraph

    :returns: NaN when :math:`\sum^B_{b=1} \pi^{*}_b <= 0`
    """
    __slots__ = ()

    _name = 'BIC'

    def calculate(self, ICm):
        B = sum(b.GetBondTypeAsDouble() for b in self.mol.GetBonds())

        with self.rethrow_zerodiv():
            log2B = np.log2(B)
            return ICm / log2B


class ComplementaryIC(TotalIC):
    r"""complementary information content descriptor.

    .. math::
        {\rm CIC}_m = \log_2 A - {\rm IC}_m

    :type order: int
    :param order: order(number of edge) of subgraph
    """
    __slots__ = ()

    _name = 'CIC'

    def calculate(self, ICm):
        A = self.mol.GetNumAtoms()

        return np.log2(A) - ICm


class ModifiedIC(InformationContent):
    r"""modified information content index descriptor.

    :type order: int
    :param order: order(number of edge) of subgraph
    """
    __slots__ = ()

    _name = 'MIC'

    def calculate(self, iAgs):
        ids, Ags = iAgs
        w = np.vectorize(lambda i: self.mol.GetAtomWithIdx(int(i)).GetMass())(ids)
        return shannon_entropy(Ags, w)


class ZModifiedIC(InformationContent):
    r"""Z-modified information content index descriptor.

    :type order: int
    :param order: order(number of edge) of subgraph
    """
    __slots__ = ()

    _name = 'ZMIC'

    def calculate(self, iAgs):
        ids, Ags = iAgs
        w = Ags * np.vectorize(lambda i: self.mol.GetAtomWithIdx(int(i)).GetAtomicNum())(ids)
        return shannon_entropy(Ags, w)



import math

from networkx import Graph

from ._base import Descriptor
from ._atomic_property import table, halogen

__all__ = ('MolecularId',)


class AtomicId(object):
    __slots__ = ('G', 'lim', 'start', 'id', 'visited', 'weights')

    def __init__(self, mol, eps):
        G = Graph()

        G.add_nodes_from(a.GetIdx() for a in mol.GetAtoms())

        for bond in mol.GetBonds():
            a = bond.GetBeginAtom()
            b = bond.GetEndAtom()

            w = a.GetDegree() * b.GetDegree()

            G.add_edge(a.GetIdx(), b.GetIdx(), weight=w)

        self.G = G
        self.lim = int(1.0 / (eps ** 2))

    def get_atomic_id(self, s):
        self.start = s
        self.id = 0.0
        self.visited = set()
        self.weights = [1]
        self._search(s)
        return self.id

    def _search(self, u):
        self.visited.add(u)

        for v, d in self.G[u].items():
            if v in self.visited:
                continue

            self.visited.add(v)
            w = d['weight'] * self.weights[-1]
            self.weights.append(w)

            self.id += 1.0 / math.sqrt(w)
            if w < self.lim:
                self._search(v)

            self.visited.remove(v)
            self.weights.pop()

    def __call__(self):
        return [
            self.get_atomic_id(i)
            for i in range(self.G.number_of_nodes())
        ]


class MolecularIdBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False
    require_connected = True

    def parameters(self):
        return self._eps,


class AtomicIds(MolecularIdBase):
    __slots__ = ('_eps',)

    def __init__(self, eps=1e-10):
        self._eps = eps

    def calculate(self):
        aid = AtomicId(self.mol, self._eps)
        return [
            1 + aid.get_atomic_id(i) / 2.0
            for i in range(self.mol.GetNumAtoms())
        ]


class MolecularId(MolecularIdBase):
    r"""molecular id descriptor.

    :type type: :py:class:`str` or :py:class:`int`
    :param type: target of atomic id source

        * 'any': normal molecular id(sum of all atomic id)
        * 'X': sum of halogen atomic id
        * str: atomic symbol
        * int: atomic number

    :type averaged: bool
    :param averaged: averaged by number of atoms

    :type _eps: float
    :param _eps: internally used
    """
    __slots__ = ('_orig_type', '_averaged', '_eps', '_type', '_check')

    @classmethod
    def preset(cls):
        return (
            cls(s, a)
            for s in ['any', 'hetero', 'C', 'N', 'O', 'X']
            for a in [False, True]
        )

    def __str__(self):
        n = 'AMID' if self._averaged else 'MID'
        if self._type != 'any':
            n = '{}_{}'.format(n, self._type)

        return n

    def parameters(self):
        return self._orig_type, self._averaged, self._eps

    def __init__(self, type='any', averaged=False, _eps=1e-10):
        self._orig_type = self._type = type
        self._averaged = averaged
        self._eps = _eps

        if isinstance(type, str) and type not in ['any', 'hetero', 'X']:
            type = table.GetAtomicNumber(type)

        if type == 'any':
            self._check = lambda _: True
        elif type == 'hetero':
            self._type = 'h'
            self._check = lambda a: a not in set([1, 6])
        elif self._type == 'X':
            self._check = lambda a: a in halogen
        else:
            self._check = lambda a: a == type

    def dependencies(self):
        return {'aids': AtomicIds(self._eps)}

    def calculate(self, aids):
        v = float(
            sum(
                aid
                for aid, atom in zip(aids, self.mol.GetAtoms())
                if self._check(atom.GetAtomicNum())
            )
        )

        if self._averaged:
            v /= self.mol.GetNumAtoms()

        return v

    rtype = float



from __future__ import division

import os

import numpy as np
from rdkit import Chem
from rdkit.Chem.rdPartialCharges import ComputeGasteigerCharges

from ._base import Descriptor
from ._util import atoms_to_numpy

halogen = set([9, 17, 35, 53, 85, 117])


getter_list = []
getters = {}


def getter(short, **attrs):
    def proc(f):
        f.short = short

        for a, v in attrs.items():
            setattr(f, a, v)

        getters[short] = f
        getter_list.append(f)
        return f

    if short in getters:
        raise ValueError('duplicated short name of atomic property')

    return proc


@getter(short='c', long='gasteiger charge', gasteiger_charges=True)
def get_gasteiger_charge(atom):
    return (
        atom.GetDoubleProp('_GasteigerCharge') +
        atom.GetDoubleProp('_GasteigerHCharge') if atom.HasProp('_GasteigerHCharge') else 0.0
    )


class PeriodicTable(object):
    __slots__ = 'data',

    _datadir = os.path.join(
        os.path.dirname(__file__),
        'data'
    )

    def __init__(self, data=None):
        self.data = data

    @classmethod
    def load(cls, name, conv=float):
        def read(v):
            if '-' in v:
                return np.nan

            try:
                return conv(v)
            except ValueError:
                return

        self = cls()

        with open(os.path.join(cls._datadir, name)) as file:
            self.data = [
                v
                for v in (read(l.split('#')[0]) for l in file)
                if v is not None
            ]

        return self

    def __getitem__(self, i):
        if i < 1:
            return np.nan

        try:
            return self.data[i - 1]
        except IndexError:
            return np.nan

    def map(self, f):
        new = self.__class__()
        new.data = [f(d) for d in self.data]
        return new

mass = PeriodicTable.load('mass.txt')
vdw_radii = PeriodicTable.load('van_der_waals_radii.txt')
vdw_volume = vdw_radii.map(lambda r: 4. / 3. * np.pi * r ** 3)
sanderson = PeriodicTable.load('sanderson_electron_negativity.txt')
pauling = PeriodicTable.load('pauling_electron_negativity.txt')
allred_rocow = PeriodicTable.load('allred_rocow_electron_negativity.txt')
polarizability94 = PeriodicTable.load('polarizalibity94.txt')
polarizability78 = PeriodicTable.load('polarizalibity78.txt')
ionization_potentials = PeriodicTable.load('ionization_potential.txt')
period = PeriodicTable(
    ([1] * 2) +
    ([2] * 8) + ([3] * 8) +
    ([4] * 18) + ([5] * 18) +
    ([6] * 32) + ([7] * 32)
)

mc_gowan_volume = PeriodicTable.load('mc_gowan_volume.txt')

table = Chem.GetPeriodicTable()


# http://dx.doi.org/10.1002%2Fjps.2600721016
@getter(short='dv', long='valence electrons', valence=True)
def get_valence_electrons(atom):
    N = atom.GetAtomicNum()
    if N == 1:
        return 0

    Zv = table.GetNOuterElecs(N) - atom.GetFormalCharge()
    Z = atom.GetAtomicNum() - atom.GetFormalCharge()
    hi = atom.GetTotalNumHs()
    he = sum(1 for a in atom.GetNeighbors() if a.GetAtomicNum() == 1)
    h = hi + he
    return (Zv - h) / (Z - Zv - 1)


@getter(short='d', long='sigma electrons', valence=True)
def get_sigma_electrons(atom):
    return sum(1 for a in atom.GetNeighbors()
               if a.GetAtomicNum() != 1)


# http://www.edusoft-lc.com/molconn/manuals/400/chaptwo.html
# p. 283
@getter(short='s', long='intrinsic state', require_connected=True, valence=True)
def get_intrinsic_state(atom):
    i = atom.GetAtomicNum()
    d = get_sigma_electrons(atom)
    dv = get_valence_electrons(atom)

    if d == 0:
        return np.nan

    return ((2. / period[i]) ** 2 * dv + 1) / d


def get_core_count(atom):
    Z = atom.GetAtomicNum()
    if Z == 1:
        return 0.0

    Zv = table.GetNOuterElecs(Z)
    PN = period[Z]

    return (Z - Zv) / (Zv * (PN - 1))


def get_eta_epsilon(atom):
    Zv = table.GetNOuterElecs(atom.GetAtomicNum())
    return 0.3 * Zv - get_core_count(atom)


def get_eta_beta_sigma(atom):
    e = get_eta_epsilon(atom)
    return sum(
        0.5 if abs(get_eta_epsilon(a) - e) <= 0.3 else 0.75
        for a in atom.GetNeighbors()
        if a.GetAtomicNum() != 1
    )


def get_eta_nonsigma_contribute(bond):
    if bond.GetBondType() is Chem.BondType.SINGLE:
        return 0.0

    f = 1.0
    if bond.GetBondTypeAsDouble() == Chem.BondType.TRIPLE:
        f = 2.0

    a = bond.GetBeginAtom()
    b = bond.GetEndAtom()

    dEps = abs(get_eta_epsilon(a) - get_eta_epsilon(b))

    if bond.GetIsAromatic():
        y = 2.0
    elif dEps > 0.3:
        y = 1.5
    else:
        y = 1.0

    return y * f


def get_eta_beta_delta(atom):
    if atom.GetIsAromatic() or\
            atom.IsInRing() or\
            table.GetNOuterElecs(atom.GetAtomicNum()) - atom.GetTotalValence() <= 0:
        return 0.0

    for b in atom.GetNeighbors():
        if b.GetIsAromatic():
            return 0.5

    return 0.0


def get_other_atom(bond, atom):
    begin = bond.GetBeginAtom()
    if atom.GetIdx() != begin.GetIdx():
        return begin

    return bond.GetEndAtom()


def get_eta_beta_non_sigma(atom):
    return sum(
        get_eta_nonsigma_contribute(b)
        for b in atom.GetBonds()
        if get_other_atom(b, atom).GetAtomicNum() != 1
    )


def get_eta_gamma(atom):
    beta = get_eta_beta_sigma(atom) + get_eta_beta_non_sigma(atom) + get_eta_beta_delta(atom)
    if beta == 0:
        return np.nan

    return get_core_count(atom) / beta


@getter(short='Z', long='atomic number')
def get_atomic_number(a):
    return a.GetAtomicNum()


@getter(short='m', long='mass')
def get_mass(a):
    return mass[a.GetAtomicNum()]


@getter(short='v', long='vdw volume')
def get_vdw_volume(a):
    return vdw_volume[a.GetAtomicNum()]


@getter(short='se', long='sanderson EN')
def get_sanderson_en(a):
    return sanderson[a.GetAtomicNum()]


@getter(short='pe', long='pauling EN')
def get_pauling_en(a):
    return pauling[a.GetAtomicNum()]


@getter(short='are', long='allred-rocow EN')
def get_allred_rocow_en(a):
    return allred_rocow[a.GetAtomicNum()]


@getter(short='p', long='polarizability')
def get_polarizability(a):
    return polarizability94[a.GetAtomicNum()]


@getter(short='i', long='ionization potential')
def get_ionization_potential(a):
    return ionization_potentials[a.GetAtomicNum()]


def get_mc_gowan_volume(a):
    return mc_gowan_volume[a.GetAtomicNum()]


def get_properties(charge=False, valence=False):
    for f in getters.values():
        if not charge and getattr(f, 'gasteiger_charges', False):
            continue

        if not valence and getattr(f, 'valence', False):
            continue

        yield f.short


class AtomicProperty(Descriptor):
    __slots__ = 'explicit_hydrogens', 'prop', '_initialized'

    def __str__(self):
        return 'Prop{}'.format(self.as_argument)

    def get_long(self):
        return getattr(self.prop, 'long', self.prop.__name__)

    @property
    def as_argument(self):
        return getattr(self.prop, 'short', self.prop.__name__)

    def parameters(self):
        return self.explicit_hydrogens, self.prop

    def __new__(cls, explicit_hydrogens, prop):
        if isinstance(prop, cls):
            prop._initialized = True
            return prop

        return super(AtomicProperty, cls).__new__(cls)

    def __init__(self, explicit_hydrogens, prop):
        if getattr(self, '_initialized', False):
            return

        self.explicit_hydrogens = explicit_hydrogens
        self.prop = getters.get(prop)

        if self.prop is not None:
            return

        if hasattr(prop, '__call__'):
            self.prop = prop
            return

        raise TypeError('atomic property is not callable: {!r}'.format(prop))

    def calculate(self):
        if getattr(self.prop, 'gasteiger_charges', False):
            ComputeGasteigerCharges(self.mol)

        r = atoms_to_numpy(self.prop, self.mol)

        nans = np.isnan(r)
        if np.any(nans):
            atms = set(np.array([a.GetSymbol() for a in self.mol.GetAtoms()])[nans])
            self.fail(ValueError('missing {} for {}'.format(self.get_long(), list(atms))))

        return r

    _carbon = Chem.Atom(6)

    @property
    def carbon(self):
        return self.prop(self._carbon)



from enum import IntEnum
from itertools import chain

from rdkit import Chem

from ._base import Descriptor
from ._util import parse_enum

__all__ = ('BondCount',)


class BondType(IntEnum):
    __slots__ = ()

    any = 1
    heavy = 2

    single = 3
    double = 4
    triple = 5

    aromatic = 6
    multiple = 7

    @property
    def as_argument(self):
        return self.name


bond_types = (
    (BondType.any, ('', lambda _: True)),
    (BondType.heavy, ('O', lambda _: True)),

    (BondType.single, ('S', lambda b: b.GetBondType() == Chem.BondType.SINGLE)),
    (BondType.double, ('D', lambda b: b.GetBondType() == Chem.BondType.DOUBLE)),
    (BondType.triple, ('T', lambda b: b.GetBondType() == Chem.BondType.TRIPLE)),

    (BondType.aromatic, ('A', lambda b: b.GetIsAromatic() or
                         b.GetBondType() == Chem.BondType.AROMATIC)),
    (BondType.multiple, ('M', lambda b: b.GetIsAromatic() or
                         b.GetBondType() != Chem.BondType.SINGLE)),
)

bond_type_dict = dict(bond_types)


class BondCount(Descriptor):
    r"""bond count descriptor.

    :type type: str
    :param type: one of bond_types

    :type kekulize: bool
    :param kekulize: use kekulized structure
    """
    __slots__ = ('_type', '_bond_name', '_check_bond', 'kekulize',)

    bond_types = tuple(b.name for b in BondType)

    @classmethod
    def preset(cls):
        return chain(
            map(lambda t: cls(t, False), BondType),
            map(lambda t: cls(t, True), [
                BondType.single, BondType.double,
            ])
        )

    def __str__(self):
        K = 'K' if self.kekulize else ''
        return 'nBonds{}{}'.format(K, self._bond_name)

    @property
    def explicit_hydrogens(self):
        return self._type in (BondType.any, BondType.single)

    def parameters(self):
        return self._type, self.kekulize

    def __init__(self, type='any', kekulize=False):
        self._type = parse_enum(BondType, type)
        self._bond_name, self._check_bond = bond_type_dict[self._type]
        self.kekulize = kekulize

    def calculate(self):
        return sum(1 for b in self.mol.GetBonds() if self._check_bond(b))

    rtype = int

    _extra_docs = 'bond_types',



from __future__ import print_function

import argparse

from rdkit import Chem

from . import SurfaceArea
from .._util import PathType, module_prog


def main():
    parser = argparse.ArgumentParser(
        prog=module_prog(__package__),
        formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.add_argument('sdf', type=PathType, help='input sd file', metavar='SDF')
    parser.add_argument('-s', '--solvent-radius', type=float, default=1.4,
                        help='solvent radius (default: %(default)s)')
    parser.add_argument('-l', '--mesh-level', type=int, default=5,
                        help='mesh level (default: %(default)s)')
    result = parser.parse_args()
    main_process(
        sdf=result.sdf, solvent_radius=result.solvent_radius, mesh_level=result.mesh_level
    )


def main_process(sdf, solvent_radius, mesh_level):
    for i, mol in enumerate(Chem.SDMolSupplier(sdf, removeHs=False)):
        name = mol.GetProp('_Name') if mol.HasProp('_Name') else str(i)
        sa = SurfaceArea.from_mol(mol, solvent_radius=solvent_radius, level=mesh_level)

        print(name)

        for a, s in zip(mol.GetAtoms(), sa.surface_area()):
            print('{:4d} {:2s} {:8.3f}'.format(
                a.GetIdx(),
                a.GetSymbol(),
                s
            ))


if __name__ == '__main__':
    main()



from __future__ import division

from collections import defaultdict

import numpy as np

from ._mesh import SphereMesh
from .._util import atoms_to_numpy
from .._atomic_property import table, vdw_radii


class SurfaceArea(object):
    r"""calculate solvent accessible surface area.

    :type radiuses: np.ndarray(dtype=float, shape=(N,))
    :param radiuses: atomic radius + solvent radius vector

    :type xyzs: np.ndarray(dtype=float, shape=(N, 3))
    :param xyzs: atomic position matrix

    :type level: int
    :param level: mesh level. subdivide icosahedron n-1 times.

        .. math::

            N_{\rm points} = 5 \times 4^{level} - 8
    """

    def __init__(self, radiuses, xyzs, level=4):
        self.rads = radiuses
        self.rads2 = radiuses ** 2
        self.xyzs = xyzs
        self._gen_neighbor_list()
        self.sphere = SphereMesh(level).vertices.T

    def _gen_neighbor_list(self):
        r = self.rads[:, np.newaxis] + self.rads

        d = np.sqrt(
            np.sum(
                (self.xyzs[:, np.newaxis] - self.xyzs) ** 2,
                axis=2
            )
        )

        ns = defaultdict(list)
        for i, j in np.transpose(np.nonzero(d <= r)):
            if i == j:
                continue

            ns[i].append((j, d[i, j]))

        for _, l in ns.items():
            l.sort(key=lambda i: i[1])

        self.neighbors = ns

    def atomic_sa(self, i):
        r"""calculate atomic surface area.

        :type i: int
        :param i: atom index

        :rtype: float
        """

        sa = 4.0 * np.pi * self.rads2[i]

        neighbors = self.neighbors.get(i)

        if neighbors is None:
            return sa

        XYZi = self.xyzs[i, np.newaxis].T

        sphere = self.sphere * self.rads[i] + XYZi
        N = sphere.shape[1]

        for j, _ in neighbors:
            XYZj = self.xyzs[j, np.newaxis].T

            d2 = (sphere - XYZj) ** 2
            mask = (d2[0] + d2[1] + d2[2]) > self.rads2[j]
            sphere = np.compress(mask, sphere, axis=1)

        return sa * sphere.shape[1] / N

    def surface_area(self):
        r"""calculate all atomic surface area.

        :rtype: [float]
        """

        return [self.atomic_sa(i) for i in range(len(self.rads))]

    @classmethod
    def from_mol(cls, mol, conformer=-1, solvent_radius=1.4, level=4):
        r"""construct SurfaceArea from rdkit Mol type.

        :type mol: rdkit.Chem.Mol
        :param mol: input molecule

        :type conformer: int
        :param conformer: conformer id

        :type solvent_radius: float
        :param solvent_radius: solvent radius

        :type level: int
        :param level: mesh level

        :rtype: SurfaceArea
        """

        rs = atoms_to_numpy(lambda a: vdw_radii[a.GetAtomicNum()] + solvent_radius, mol)

        conf = mol.GetConformer(conformer)

        ps = np.array([list(conf.GetAtomPosition(i)) for i in range(mol.GetNumAtoms())])

        return cls(rs, ps, level)

    @classmethod
    def from_pdb(cls, pdb, solvent_radius=1.4, level=3):
        try:
            from Bio.PDB import PDBParser
        except ImportError:
            raise ImportError("There isn't biopython package.")

        rs = []
        coords = []

        for atom in PDBParser().get_structure('', pdb).get_atoms():
            rs.append(vdw_radii[table.GetAtomicNumber(atom.element)] + solvent_radius)
            coords.append(atom.coord)

        return cls(np.array(rs), np.array(coords), level)



from math import pi

from ._base import Descriptor
from .BondCount import BondCount
from .RingCount import RingCount
from ._atomic_property import table

__all__ = (
    'VdwVolumeABC',
)


bondi_radii = {
    'H': 1.20,
    'C': 1.70,
    'N': 1.55,
    'O': 1.52,
    'F': 1.47,
    'Cl': 1.75,
    'Br': 1.85,
    'P': 1.80,
    'S': 1.80,
    'As': 1.85,
    'B': 2.13,
    'Si': 2.10,
    'Se': 1.90,
}


atom_contrib = {
    table.GetAtomicNumber(s): 4. / 3. * pi * r ** 3
    for s, r in bondi_radii.items()
}


class VdwVolumeABC(Descriptor):
    r"""van der waals volume(ABC) descriptor.

    :returns: NaN when any atoms are non-compat_atoms

    References
        * :doi:`10.1021/jo034808o`
    """
    __slots__ = ()

    compat_atoms = tuple(bondi_radii)

    @classmethod
    def preset(cls):
        yield cls()

    def __str__(self):
        return 'Vabc'

    def parameters(self):
        return ()

    def dependencies(self):
        return {
            'Nb': BondCount(),
            'NRa': RingCount(None, False, False, True, None),
            'NRA': RingCount(None, False, False, False, None),
        }

    def calculate(self, Nb, NRa, NRA):
        try:
            ac = sum(
                atom_contrib[a.GetAtomicNum()]
                for a in self.mol.GetAtoms()
            )
        except KeyError:
            self.fail(ValueError('unknown atom type'))

        return ac - 5.92 * Nb - 14.7 * NRa - 3.8 * NRA

    rtype = float



from rdkit import Chem
from numpy.testing import assert_almost_equal

from mordred.SLogP import SMR, SLogP


def test_WildmanCrippen1():
    mol = Chem.MolFromSmiles('Oc1ccccc1OC')
    yield assert_almost_equal, SLogP()(mol), 1.4, 2
    yield assert_almost_equal, SMR()(mol), 34.66, 2


def test_WildmanCrippen2():
    mol = Chem.MolFromSmiles('c1ccccc1c2ccccn2')
    yield assert_almost_equal, SLogP()(mol), 2.75, 2



r"""modred descriptor calculator."""

from ._version import __version__

from ._base import (
    all_descriptors,
    Calculator,
    Descriptor,
    get_descriptors_from_module,
)

__all__ = (
    '__version__',
    'Descriptor',
    'Calculator',
    'all_descriptors',
    'get_descriptors_from_module',
)



from ._base import Descriptor
from .SLogP import SMR, SLogP
from .Weight import Weight
from .HydrogenBond import HBondDonor, HBondAcceptor

__all__ = ('Lipinski', 'GhoseFilter')


class LipinskiLike(Descriptor):
    __slots__ = ()

    @classmethod
    def preset(cls):
        yield cls()

    def parameters(self):
        return ()

    def __str__(self):
        return self.__class__.__name__

    rtype = bool


class Lipinski(LipinskiLike):
    r"""Lipinski rule of 5 descriptor.

    LogP: SLogP
    """
    __slots__ = ()

    def dependencies(self):
        return {
            'LogP': SLogP(),
            'MW': Weight(),
            'HBDon': HBondDonor(),
            'HBAcc': HBondAcceptor(),
        }

    def calculate(self, LogP, MW, HBDon, HBAcc):
        return\
            HBDon <= 5 and\
            HBAcc <= 10 and\
            MW <= 500 and\
            LogP <= 5


class GhoseFilter(LipinskiLike):
    r"""Ghose filter descriptor.

    LogP, MR: SLogP, SMR
    """
    __slots__ = ()

    def dependencies(self):
        return {
            'LogP': SLogP(),
            'MR': SMR(),
            'MW': Weight(),
        }

    def calculate(self, MW, LogP, MR):
        return\
            (160 <= MW <= 480) and\
            (20 <= self.mol.GetNumAtoms() <= 70) and\
            (-0.4 <= LogP <= 5.6) and\
            (40 <= MR <= 130)



'''
References:

    * http://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/AcidicGroupCountDescriptor.html
    * http://cdk.github.io/cdk/1.5/docs/api/org/openscience/cdk/qsar/descriptors/molecular/BasicGroupCountDescriptor.html
'''

from abc import abstractproperty

from rdkit import Chem

from ._base import Descriptor

__all__ = ('AcidicGroupCount', 'BasicGroupCount',)


class SmartsCountBase(Descriptor):
    __slots__ = '_mol',

    @classmethod
    def preset(cls):
        yield cls()

    def _create_smarts(self):
        s = ','.join('$(' + s + ')' for s in self.SMARTS)
        self._mol = Chem.MolFromSmarts('[' + s + ']')
        return self._mol

    @abstractproperty
    def SMARTS(self):
        pass

    def __str__(self):
        return self._name

    def parameters(self):
        return ()

    def calculate(self):
        pat = getattr(self, '_mol', None) or self._create_smarts()
        return len(self.mol.GetSubstructMatches(pat))

    rtype = int

    _extra_docs = 'SMARTS',


class AcidicGroupCount(SmartsCountBase):
    r"""acidic group count descriptor."""

    __slots__ = ()

    _name = 'nAcid'

    SMARTS = (
        '[O;H1]-[C,S,P]=O',
        '[*;-;!$(*~[*;+])]',
        '[NH](S(=O)=O)C(F)(F)F',
        'n1nnnc1',
    )


class BasicGroupCount(SmartsCountBase):
    r"""basic group count descriptor."""

    __slots__ = ()

    _name = 'nBase'

    SMARTS = (
        '[NH2]-[CX4]',
        '[NH](-[CX4])-[CX4]',
        'N(-[CX4])(-[CX4])-[CX4]',
        '[*;+;!$(*~[*;-])]',
        'N=C-N',
        'N-C=N',
    )



from __future__ import division

from ._base import Descriptor
from ._graph_matrix import Radius as CRadius
from ._graph_matrix import Diameter as CDiameter

__all__ = ('Diameter', 'Radius', 'TopologicalShapeIndex', 'PetitjeanIndex',)


class TopologicalIndexBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False

    @classmethod
    def preset(cls):
        yield cls()

    def parameters(self):
        return ()

    rtype = int


class Radius(TopologicalIndexBase):
    r"""radius descriptor."""
    __slots__ = ()

    def __str__(self):
        return 'Radius'

    def dependencies(self):
        return {'R': CRadius(self.explicit_hydrogens)}

    def calculate(self, R):
        return int(R)


class Diameter(TopologicalIndexBase):
    r"""diameter descriptor."""
    __slots__ = ()

    def __str__(self):
        return 'Diameter'

    def dependencies(self):
        return {'D': CDiameter(self.explicit_hydrogens)}

    def calculate(self, D):
        return int(D)


class TopologicalShapeIndex(TopologicalIndexBase):
    r"""topological shape index descriptor.

    .. math::

        I_{\rm topo} = \frac{D - R}{R}

    where
    :math:`R` is graph radius,
    :math:`D` is graph diameter.

    :returns: NaN when :math:`R = 0`
    """
    __slots__ = ()

    def __str__(self):
        return 'TopoShapeIndex'

    def dependencies(self):
        return {
            'R': CRadius(self.explicit_hydrogens),
            'D': CDiameter(self.explicit_hydrogens),
        }

    def calculate(self, R, D):
        with self.rethrow_zerodiv():
            return (D - R) / R

    rtype = float


class PetitjeanIndex(TopologicalShapeIndex):
    r"""Petitjean index descriptor.

    .. math::

        I_{\rm Petitjean} = \frac{D - R}{D}

    where
    :math:`R` is graph radius,
    :math:`D` is graph diameter.

    :returns: NaN when :math:`D = 0`
    """
    __slots__ = ()

    def __str__(self):
        return 'PetitjeanIndex'

    def calculate(self, R, D):
        with self.rethrow_zerodiv():
            return (D - R) / D



from __future__ import print_function

import os
import sys
import logging
import argparse
from importlib import import_module
from multiprocessing import freeze_support

from rdkit import Chem

from . import Calculator, __version__, descriptors
from ._base import get_descriptors_from_module
from ._util import PathType, module_prog
from .error import Missing, MissingValueBase


def smiles_parser(path):
    with open(path) as file:
        for line in file:
            line = line.strip().split()
            if len(line) == 1:
                smi = line[0]
                name = smi
            else:
                smi = line[0]
                name = ' '.join(line[1:])

            mol = Chem.MolFromSmiles(smi)

            if mol is None:
                logging.warning('smiles read failure: %s', name)
                continue

            mol.SetProp('_Name', name)
            yield mol


def sdf_parser(path):
    base = os.path.splitext(os.path.basename(path))[0]

    for i, mol in enumerate(Chem.SDMolSupplier(path, removeHs=False)):
        if mol is None:
            logging.warning('mol read failure: %s.%s', base, i)
            continue

        if mol.GetProp('_Name') == '':
            mol.SetProp(
                '_Name',
                '{}.{}'.format(base, i)
            )

        yield mol


def auto_parser(path):
    ext = os.path.splitext(path)[1]
    if ext == '.smi':
        r = smiles_parser(path)
    elif ext in ['.mol', '.sdf']:
        r = sdf_parser(path)
    else:
        logging.warning('cannot detect file format: %s', path)
        r = ()

    for m in r:
        yield m


class ParserAction(argparse.Action):
    def __init__(self, option_strings, dest, **kwargs):
        super(ParserAction, self).__init__(option_strings, dest, **kwargs)
        self.default = self.to_parser('auto')
        self.choices = ['auto', 'sdf', 'mol', 'smi']

    def to_parser(self, value):
        if value == 'auto':
            return auto_parser
        elif value == 'smi':
            return smiles_parser
        elif value in ['sdf', 'mol']:
            return sdf_parser

        raise ValueError('invalid parser: {}'.format(value))

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self.to_parser(values))


def make_parser():
    parser = argparse.ArgumentParser(
        prog=module_prog(__package__),
        epilog='descriptors: {}'.format(' '.join(descriptors.__all__))
    )
    parser.add_argument('--version', action='version', help='input molecular file',
                        version='{}-{}'.format(__package__, __version__))
    parser.add_argument('input', type=PathType, nargs='+', metavar='INPUT')
    parser.add_argument('-t', '--type', action=ParserAction,
                        help='input filetype (default: auto)')
    parser.add_argument('-o', '--output', default='-', type=argparse.FileType('w'),
                        help='output file path (default: stdout)')
    parser.add_argument('-p', '--processes', default=None, type=int,
                        help='number of processes (default: number of logical processors)')
    parser.add_argument('-q', '--quiet', action='store_true', help='hide progress bar')
    parser.add_argument('-s', '--stream', action='store_true', help='stream read')
    parser.add_argument('-d', '--descriptor', default=[], choices=descriptors.__all__, action='append',
                        help='descriptors to calculate (default: all)', metavar='DESC')
    parser.add_argument('-3', '--3D', action='store_true', dest='with3D',
                        help='use 3D descriptors (require sdf or mol file)')
    parser.add_argument('-v', '--verbosity', action='count', default=0, help='verbosity')

    return parser


def main_process(input, parser, output, nproc, quiet, stream, descriptor, with3D, verbosity):
    mols = (m for i in input for m in parser(i))

    if output.isatty():
        quiet = True

    if stream:
        N = None
    else:
        mols = list(mols)
        N = len(mols)

    # Descriptors
    calc = Calculator()

    if verbosity >= 2:
        calc._debug = True

    if len(descriptor) == 0:
        calc.register(descriptors, ignore_3D=not with3D)
    else:
        calc.register(
            (d
             for m in descriptor
             for d in get_descriptors_from_module(import_module('.' + m, __package__))
             ),
            ignore_3D=not with3D
        )

    with output:
        write_row(output, ['name'] + [str(d) for d in calc.descriptors])

        def warning(name, v, err_set):
            if not isinstance(v, MissingValueBase):
                return

            if verbosity == 0 and isinstance(v, Missing):
                return

            red = v.error.__class__, v.error.args
            if red in err_set:
                return

            calc.echo('[{}] {}: {}'.format(v.header, name, v), file=sys.stderr)
            err_set.add(red)

        def pretty(name, v, err_set):
            warning(name, v, err_set)

            if isinstance(v, MissingValueBase):
                return ''

            return str(v)

        for mol, val in zip(mols, calc.map(mols, nproc=nproc, nmols=N, quiet=quiet)):
            err_set = set()

            if mol.HasProp('_Name'):
                name = mol.GetProp('_Name')
            else:
                name = Chem.MolToSmiles(mol)

            write_row(output, [name] + [pretty(name, v, err_set) for v in val])


def write_row(file, data):
    file.write(
        ','.join(
            str(v).replace('"', '""').replace('\n', '').replace('\r', '')
            for v in data
        )
    )
    file.write('\n')


def main(args=None):
    parser = make_parser()
    p = parser.parse_args(args)
    return main_process(
        input=p.input, parser=p.type, output=p.output,
        nproc=p.processes, quiet=p.quiet, stream=p.stream,
        descriptor=p.descriptor, with3D=p.with3D, verbosity=p.verbosity
    )


if __name__ == '__main__':
    freeze_support()
    main()



from setuptools import setup, find_packages
import sys
import os
import io


install_requires = [
    'six>=1.10',
    'numpy>=1.10',
    'networkx>=1.10',
    'tqdm>=3.7.1',
]

if sys.version_info < (3, 4, 0):
    install_requires.append('enum34')


def get_version():
    with open(os.path.join('mordred', '_version.txt')) as f:
        return f.read().strip()


def get_test_data():
    for p, _, fs in os.walk(os.path.join('mordred', 'tests', 'references')):
        p = p.split(os.sep)[2:]

        for f in fs:
            yield os.path.join(*(p + [f]))

README_rst = ''
fndoc = os.path.join(os.path.dirname(__file__), 'README.rst')
with io.open(fndoc, mode='r', encoding='utf-8') as fd:
    README_rst = fd.read()

setup(
    name='mordred',
    version=get_version(),
    description='molecular descriptor calculator',
    long_description=README_rst,
    license='BSD3',
    author='Hirotomo Moriwaki',
    author_email='philopon.dependence@gmail.com',
    url='https://github.com/mordred-descriptor/mordred',
    platforms=['any'],
    keywords='QSAR chemoinformatics',

    packages=find_packages(),

    package_data={
        'mordred': ['data/*.txt', '_version.txt'],
        'mordred.tests': list(get_test_data()),
    },

    install_requires=install_requires,

    tests_require=[
        'nose>=1.3',
        'PyYaml>=3.11',
    ],

    test_suite='nose.collector',
)



from ._base import Descriptor

__all__ = (
    'AromaticAtomsCount', 'AromaticBondsCount',
)


class AromaticBase(Descriptor):
    __slots__ = ()

    @classmethod
    def preset(cls):
        yield cls()

    def __str__(self):
        return self._name

    def parameters(self):
        return ()

    rtype = int


class AromaticAtomsCount(AromaticBase):
    r"""aromatic atoms count descriptor."""
    __slots__ = ()
    _name = 'nAromAtom'

    def calculate(self):
        return sum(
            1
            for a in self.mol.GetAtoms()
            if a.GetIsAromatic()
        )


class AromaticBondsCount(AromaticBase):
    r"""aromatic bonds count descriptor."""
    __slots__ = ()
    _name = 'nAromBond'

    def calculate(self):
        return sum(
            1
            for b in self.mol.GetBonds()
            if b.GetIsAromatic()
        )



import math
import pickle

import six
from rdkit import Chem
from nose.tools import eq_
from numpy.testing import assert_almost_equal

from mordred import Calculator, descriptors
from mordred.error import MissingValueBase


def test_pickle_calculator():
    orig = Calculator(descriptors)
    d0 = orig.descriptors[0]
    d1 = orig.descriptors[1]
    orig.register([
        d0 + d1, d0 - d1, d0 * d1, d0 // d1, d0 % d1, d0 ** d1,
        -d0, +d1, abs(d0), math.trunc(d0),
    ])

    if six.PY3:
        orig.register([math.ceil(d0), math.floor(d1)])

    pickled = pickle.loads(pickle.dumps(orig))

    mol = Chem.MolFromSmiles('c1ccccc1C(O)O')

    for a, b in zip(orig.descriptors, pickled.descriptors):
        yield eq_, a, b

    for a, b in zip(orig(mol), pickled(mol)):
        if isinstance(a, MissingValueBase):
            yield eq_, a.__class__, b.__class__
        else:
            yield assert_almost_equal, a, b
